# QCoDeS driver for QDac using channels

import logging
import time
from functools import partial
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import pyvisa as visa
from pyvisa.resources.serial import SerialInstrument

from qcodes.instrument.base import Instrument
from qcodes.instrument.channel import (
    ChannelList,
    InstrumentChannel,
    MultiChannelInstrumentParameter,
)
from qcodes.instrument.parameter import ParamRawDataType
from qcodes.instrument.visa import VisaInstrument
from qcodes.utils import validators as vals

log = logging.getLogger(__name__)


class QDacChannel(InstrumentChannel):
    """
    A single output channel of the QDac.

    Exposes chan.v, chan.vrange, chan.slope, chan.i, chan.irange
    """

    _CHANNEL_VALIDATION = vals.Numbers(1, 48)

    def __init__(self, parent: Instrument, name: str, channum: int):
        """
        Args:
            parent: The instrument to which the channel is
                attached.
            name: The name of the channel
            channum: The number of the channel in question (1-48)
        """
        super().__init__(parent, name)

        # Validate the channel
        self._CHANNEL_VALIDATION.validate(channum)

        # Add the parameters

        self.add_parameter('v',
                           label=f'Channel {channum} voltage',
                           unit='V',
                           set_cmd=partial(self._parent._set_voltage, channum),
                           get_cmd=partial(self._parent._get_voltage, channum),
                           get_parser=float,
                           vals=vals.Numbers(-10, 10)
                           )

        self.add_parameter('vrange',
                           label=f'Channel {channum} atten.',
                           set_cmd=partial(self._parent._set_vrange, channum),
                           get_cmd=partial(self._parent._get_vrange, channum),
                           vals=vals.Enum(0, 1)
                           )

        self.add_parameter('i',
                           label=f'Channel {channum} current',
                           get_cmd=f'get {channum}',
                           unit='A',
                           get_parser=self._parent._current_parser
                           )

        self.add_parameter('irange',
                           label=f'Channel {channum} irange',
                           set_cmd=f'cur {channum} {{}}',
                           get_cmd=f'cur {channum}',
                           get_parser=int
                           )

        self.add_parameter('slope',
                           label=f'Channel {channum} slope',
                           unit='V/s',
                           set_cmd=partial(self._parent._setslope, channum),
                           get_cmd=partial(self._parent._getslope, channum),
                           vals=vals.MultiType(vals.Enum('Inf'),
                                               vals.Numbers(1e-3, 100))
                           )

        self.add_parameter('sync',
                           label=f'Channel {channum} sync output',
                           set_cmd=partial(self._parent._setsync, channum),
                           get_cmd=partial(self._parent._getsync, channum),
                           vals=vals.Ints(0, 5)
                           )

        self.add_parameter(name='sync_delay',
                           label=f'Channel {channum} sync pulse delay',
                           unit='s',
                           get_cmd=None, set_cmd=None,
                           initial_value=0
                           )

        self.add_parameter(name='sync_duration',
                           label=f'Channel {channum} sync pulse duration',
                           unit='s',
                           get_cmd=None, set_cmd=None,
                           initial_value=0.01
                           )

    def snapshot_base(
            self,
            update: Optional[bool] = False,
            params_to_skip_update: Optional[Sequence[str]] = None
    ) -> Dict[Any, Any]:
        update_currents = self._parent._update_currents and update
        if update and not self._parent._get_status_performed:
            self._parent._update_cache(readcurrents=update_currents)
        # call get_status rather than getting the status individually for
        # each parameter. This is only done if _get_status_performed is False
        # this is used to signal that the parent has already called it and
        # no need to repeat.
        if params_to_skip_update is None:
            params_to_skip_update = ('v', 'i', 'irange', 'vrange')
        snap = super().snapshot_base(update=update,
                                     params_to_skip_update=params_to_skip_update)
        return snap


class QDacMultiChannelParameter(MultiChannelInstrumentParameter):
    """
    The class to be returned by __getattr__ of the ChannelList. Here customised
    for fast multi-readout of voltages.
    """
    def __init__(
            self,
            channels: Sequence[InstrumentChannel],
            param_name: str,
            *args: Any,
            **kwargs: Any):
        super().__init__(channels, param_name, *args, **kwargs)

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        """
        Return a tuple containing the data from each of the channels in the
        list.
        """
        # For voltages, we can do something slightly faster than the naive
        # approach

        if self._param_name == 'v':
            qdac = self._channels[0]._parent
            qdac._update_cache(readcurrents=False)
            output = tuple(chan.parameters[self._param_name].get_latest()
                           for chan in self._channels)
        else:
            output = tuple(chan.parameters[self._param_name].get()
                           for chan in self._channels)

        return output


class QDac(VisaInstrument):
    """
    Channelised driver for the QDev digital-analog converter QDac

    Based on "DAC_commands_v_13.pdf"
    Tested with Software Version: 0.170202

    The driver assumes that the instrument is ALWAYS in verbose mode OFF
    """


    # set nonzero value (seconds) to accept older status when reading settings
    max_status_age = 1

    def __init__(self,
                 name: str,
                 address: str,
                 num_chans: int = 48,
                 update_currents: bool = True,
                 **kwargs: Any):
        """
        Instantiates the instrument.

        Args:
            name: The instrument name used by qcodes
            address: The VISA name of the resource
            num_chans: Number of channels to assign. Default: 48
            update_currents: Whether to query all channels for their
                current current value on startup. Default: True.

        Returns:
            QDac object
        """
        super().__init__(name, address, **kwargs)
        self._output_n_lines = 50
        handle = self.visa_handle
        assert isinstance(handle, SerialInstrument)
        self._get_status_performed = False
        # This is the baud rate on power-up. It can be changed later but
        # you must start out with this value.
        handle.baud_rate = 480600
        handle.parity = visa.constants.Parity(0)
        handle.data_bits = 8
        self.set_terminator('\n')
        # TODO: do we want a method for write termination too?
        handle.write_termination = '\n'
        # TODO: do we need a query delay for robust operation?
        self._write_response = ''

        if self._get_firmware_version() < 0.170202:
            raise RuntimeError('''
                               Obsolete QDAC Software version detected.
                               QCoDeS only supports version 0.170202 or newer.
                               Contact rikke.lutge@nbi.ku.dk for an update.
                               ''')

        self.num_chans = num_chans

        # Assigned slopes. Entries will eventually be [chan, slope]
        self._slopes: List[Tuple[int, Union[str, float]]] = []
        # Function generators (used in _set_voltage)
        self._fgs = set(range(1, 9))
        self._assigned_fgs: Dict[int, int] = {}  # {chan: fg}
        # Sync channels
        self._syncoutputs: List[Tuple[int, int]] = []  # Entries: [chan, syncchannel]

        self.chan_range = range(1, 1 + self.num_chans)
        self.channel_validator = vals.Ints(1, self.num_chans)

        channels = ChannelList(self, "Channels", QDacChannel,
                               snapshotable=False,
                               multichan_paramclass=QDacMultiChannelParameter)

        for i in self.chan_range:
            channel = QDacChannel(self, f'chan{i:02}', i)
            channels.append(channel)
            # Should raise valueerror if name is invalid (silently fails now)
            self.add_submodule(f'ch{i:02}', channel)
        channels.lock()
        self.add_submodule('channels', channels)

        for board in range(6):
            for sensor in range(3):
                label = f'Board {board}, Temperature {sensor}'
                self.add_parameter(name=f'temp{board}_{sensor}',
                                   label=label,
                                   unit='C',
                                   get_cmd=f'tem {board} {sensor}',
                                   get_parser=self._num_verbose)

        self.add_parameter(name='cal',
                           set_cmd='cal {}',
                           vals=self.channel_validator)
        # TO-DO: maybe it's too dangerous to have this settable.
        # And perhaps ON is a better verbose mode default?
        self.add_parameter(name='verbose',
                           set_cmd='ver {}',
                           val_mapping={True: 1, False: 0})

        # Initialise the instrument, all channels DC (unbind func. generators)
        for chan in self.chan_range:
            # Note: this call does NOT change the voltage on the channel
            self.write(f'wav {chan} 0 1 0')

        self.verbose.set(False)
        self.connect_message()
        log.info('[*] Querying all channels for voltages and currents...')
        self._update_cache(readcurrents=update_currents)
        self._update_currents = update_currents
        log.info('[+] Done')

    def snapshot_base(
            self,
            update: Optional[bool] = False,
            params_to_skip_update: Optional[Sequence[str]] = None
    ) -> Dict[Any, Any]:
        update_currents = self._update_currents and update is True
        if update:
            self._update_cache(readcurrents=update_currents)
            self._get_status_performed = True
        # call get_status rather than getting the status individually for
        # each parameter. We set _get_status_performed to True
        # to indicate that each update channel does not need to call this
        # function as opposed to when snapshot is called on an individual
        # channel
        snap = super().snapshot_base(update=update,
                                     params_to_skip_update=params_to_skip_update)
        self._get_status_performed = False
        return snap

    #########################
    # Channel gets/sets
    #########################

    def _set_voltage(
            self,
            chan: int,
            v_set: float) -> None:
        """
        set_cmd for the chXX_v parameter

        Args:
            chan: The 1-indexed channel number
            v_set: The target voltage

        If a finite slope has been assigned, we assign a function generator to
        ramp the voltage.
        """
        channel = self.channels[chan-1]

        slopechans = [sl[0] for sl in self._slopes]
        if chan in slopechans:
            slope = [sl[1] for sl in self._slopes if sl[0] == chan][0]
            # find and assign fg
            fg = min(self._fgs.difference(set(self._assigned_fgs.values())))
            self._assigned_fgs[chan] = fg
            # We need .get and not get_latest in case a ramp was interrupted
            v_start = channel.v.get()
            time = abs(v_set-v_start)/slope
            log.info(f'Slope: {slope}, time: {time}')
            # Attenuation compensation and syncing
            # happen inside _rampvoltage
            self._rampvoltage(chan, fg, v_start, v_set, time)
        else:
            v_dac = QDac._get_v_dac_from_v_exp(channel, v_set)
            # set the mode back to DC in case it had been changed
            # and then set the voltage
            self.write(f'wav {chan} 0 0 0;set {chan} {v_dac:.6f}')

    def _get_voltage(self, chan: int) -> float:
        """
        get_cmd for the chXX_v parameter

        Args:
            chan: The 1-indexed channel number
        """
        self._update_cache(readcurrents=False)
        return self.channels[chan - 1].v.cache()


    # In order to get conversions right let us define a vocabulary:
    # v_exp: is the voltage including the attenuation.This is the value
    # we want to store in the cache and the value we interact with as a
    # qcodes user.
    # v_dac: this is the voltage generated by the dac and handled by the VISA
    # commands.
    # Then we have the general relationship`v_exp = v_dac * attenuation`,
    @staticmethod
    def _get_attenuation(channel: QDacChannel) -> float:
        return 0.1 if channel.vrange.cache() == 1 else 1.0

    @staticmethod
    def _get_v_dac_from_v_exp(channel: QDacChannel, v_exp: float) -> float:
        return v_exp / QDac._get_attenuation(channel)

    @staticmethod
    def _get_v_exp_from_v_dac(channel: QDacChannel, v_dac: float) -> float:
        return v_dac * QDac._get_attenuation(channel)

    def _set_vrange(self, chan: int, switchint: int) -> None:
        """
        set_cmd for the chXX_vrange parameter

        The switchint is an integer. 1 means attenuation ON.

        Since the vrange is actually a 20 dB attenuator (amplitude factor 0.1)
        immediately applied to the channel output, we must update the voltage
        parameter accordingly
        """

        self.write(f'vol {chan} {switchint}')

        # setting v_range preserves v_dac but changes v_exp, see comment above
        # for definitions.
        channel = self.channels[chan-1]
        if channel.vrange.cache() != switchint:
            v_dac = QDac._get_v_dac_from_v_exp(channel, channel.v.cache())
            channel.vrange.cache.set(switchint)
            self._update_v_validator(channel, switchint)
            channel.v.cache.set(QDac._get_v_exp_from_v_dac(channel, v_dac))

    def _get_vrange(self, chan: int) -> float:
        """
        get_cmd for the chXX_vrange parameter

        Args:
            chan: The 1-indexed channel number
        """
        self._update_cache(readcurrents=False)
        return self.channels[chan - 1].vrange.cache()

    def _num_verbose(self, s: str) -> float:
        """
        turn a return value from the QDac into a number.
        If the QDac is in verbose mode, this involves stripping off the
        value descriptor.
        """
        if self.verbose.get_latest():
            s = s.split(': ')[-1]
        return float(s)

    def _current_parser(self, s: str) -> float:
        """
        parser for chXX_i parameter
        """
        return 1e-6*self._num_verbose(s)

    def _update_cache(self, readcurrents: bool = False) -> None:
        r"""
        Function to query the instrument and get the status of all channels,
        e.g. voltage (``v``), voltage range (``vrange``), and current range (``irange``)
        parameters of all the channels.
        Takes a while to finish.

        The `status` call generates 51 lines of output. Send the command and
        read the first one, which is the software version line
        the full output looks like:
        Software Version: 0.160218\r\n
        Channel\tOut V\t\tVoltage range\tCurrent range\n
        \n
        8\t  0.000000\t\tX 1\t\tpA\n
        7\t  0.000000\t\tX 1\t\tpA\n
        ... (all 48 channels like this in a somewhat peculiar order)
        (no termination afterward besides the \n ending the last channel)
        returns a list of dicts [{v, vrange, irange}]
        NOTE - channels are 1-based, but the return is a list, so of course
        0-based, ie chan1 is out[0]
        """

        def validate_version(version_line: str) -> None:
            if version_line.startswith('Software Version: '):
                self.version = version_line.strip().split(': ')[1]
            else:
                self._wait_and_clear()
                raise ValueError('unrecognized version line: ' + version_line)

        def validate_header(header_line: str) -> None:
            headers = header_line.lower().strip('\r\n').split('\t')
            expected_headers = ['channel', 'out v', '', 'voltage range',
                                'current range']
            if headers != expected_headers:
                raise ValueError('unrecognized header line: ' + header_line)

        def parse_line(line: str) -> Tuple[int, int, int, float]:
            i_range_trans = {'hi cur': 1, 'lo cur': 0}
            v_range_trans = {'X 1': 0, 'X 0.1': 1}

            chan_str, v_str, _, v_range_str, _, i_range_str = line.split('\t')
            chan = int(chan_str)
            v_dac = float(v_str)
            v_range = v_range_trans[v_range_str.strip()]
            i_range = i_range_trans[i_range_str.strip()]
            return chan, i_range, v_range, v_dac

        validate_version(self.ask('status'))
        validate_header(self.read())

        chans_left = set(self.chan_range)
        while chans_left:
            line = self.read().strip()
            if not line:
                continue
            chan, i_range, v_range, v_dac = parse_line(line)

            channel = self.channels[chan - 1]
            channel.vrange.cache.set(v_range)
            self._update_v_validator(channel, v_range)
            channel.irange.cache.set(i_range)
            channel.v.cache.set(QDac._get_v_exp_from_v_dac(channel, v_dac))

            chans_left.remove(chan)

        if readcurrents:
            self._read_currents()

    def _read_currents(self) -> None:
        for chan in range(1, self.num_chans + 1):
            param = self.channels[chan - 1].i
            _ = param.get()

    @staticmethod
    def _update_v_validator(channel: QDacChannel, v_range: int) -> None:
        range = (-10.01, 10.01) if v_range == 0 else (-1.001, 1.001)
        channel.v.vals = vals.Numbers(*range)

    def _setsync(self, chan: int, sync: int) -> None:
        """
        set_cmd for the chXX_sync parameter.

        Args:
            chan (int): The channel number (1-48)
            sync (int): The associated sync output. 0 means 'unassign'
        """

        if chan not in range(1, 49):
            raise ValueError('Channel number must be 1-48.')

        if sync == 0:
            # try to remove the sync from internal bookkeeping
            try:
                sc = self._syncoutputs
                to_remove = [sc.index(syn) for syn in sc if syn[0] == chan][0]
                self._syncoutputs.remove(sc[to_remove])
            except IndexError:
                pass
            # free the previously assigned sync
            oldsync = self.channels[chan-1].sync.get_latest()
            if oldsync is not None:
                self.write(f'syn {oldsync} 0 0 0')
            return

        if sync in [syn[1] for syn in self._syncoutputs]:
            oldchan = [syn[0] for syn in self._syncoutputs if syn[1] == sync][0]
            self._syncoutputs.remove((oldchan, sync))

        if chan in [syn[0] for syn in self._syncoutputs]:
            oldsyn = [syn[1] for syn in self._syncoutputs if syn[0] == chan][0]
            self._syncoutputs[
                self._syncoutputs.index((chan, oldsyn))
            ] = (chan, sync)
            return

        self._syncoutputs.append((chan, sync))
        return

    def _getsync(self, chan: int) -> float:
        """
        get_cmd of the chXX_sync parameter
        """
        if chan in [syn[0] for syn in self._syncoutputs]:
            sync = [syn[1] for syn in self._syncoutputs if syn[0] == chan][0]
            return sync
        else:
            return 0

    def _setslope(self, chan: int, slope: Union[float, str]) -> None:
        """
        set_cmd for the chXX_slope parameter, the maximum slope of a channel.

        Args:
            chan: The channel number (1-48)
            slope: The slope in V/s. Write 'Inf' to allow
              arbitrary small rise times.
        """
        if chan not in range(1, 49):
            raise ValueError('Channel number must be 1-48.')

        if slope == 'Inf':
            self.write(f'wav {chan} 0 0 0')

            # Now clear the assigned slope and function generator (if possible)
            try:
                self._assigned_fgs.pop(chan)
            except KeyError:
                pass
            # Remove a sync output, if one was assigned
            syncchans = [syn[0] for syn in self._syncoutputs]
            if chan in syncchans:
                self.channels[chan-1].sync.set(0)
            try:
                sls = self._slopes
                to_remove = [sls.index(sl) for sl in sls if sl[0] == chan][0]
                self._slopes.remove(sls[to_remove])
                return
            # If the value was already 'Inf', the channel was not
            # in the list and nothing happens
            except IndexError:
                return

        if chan in [sl[0] for sl in self._slopes]:
            oldslope = [sl[1] for sl in self._slopes if sl[0] == chan][0]
            self._slopes[self._slopes.index((chan, oldslope))] = (chan, slope)
            return

        if len(self._slopes) >= 8:
            rampchans = ", ".join(str(c[0]) for c in self._slopes)
            raise ValueError(
                "Can not assign finite slope to more than "
                + "8 channels. Assign 'Inf' to at least one of "
                + f"the following channels: {rampchans}"
            )

        self._slopes.append((chan, slope))
        return

    def _getslope(self, chan: int) -> Union[str, float]:
        """
        get_cmd of the chXX_slope parameter
        """
        if chan in [sl[0] for sl in self._slopes]:
            slope = [sl[1] for sl in self._slopes if sl[0] == chan][0]
            return slope
        else:
            return 'Inf'

    def printslopes(self) -> None:
        """
        Print the finite slopes assigned to channels
        """
        for sl in self._slopes:
            print(f"Channel {sl[0]}, slope: {sl[1]} (V/s)")

    def _rampvoltage(
            self,
            chan: int,
            fg: int,
            v_start: float,
            setvoltage: float,
            ramptime: float
    ) -> None:
        """
        Smoothly ramp the voltage of a channel by the means of a function
        generator. Helper function used by _set_voltage.

        Args:
            chan: The channel number (counting from 1)
            fg: The function generator (counting from 1)
            setvoltage: The voltage to ramp to
            ramptime: The ramp time in seconds.
        """

        # Crazy stuff happens if the period is too small, e.g. the channel
        # can jump to its max voltage
        if ramptime <= 0.002:
            ramptime = 0
            log.warning('Cancelled a ramp with a ramptime of '
                        '{} s'.format(ramptime) + '. Voltage not changed.')

        offset = v_start
        amplitude = setvoltage-v_start
        if self.channels[chan-1].vrange.get_latest() == 1:
            offset *= 10
            amplitude *= 10

        chanmssg = 'wav {} {} {} {}'.format(chan, fg,
                                            amplitude,
                                            offset)

        if chan in [syn[0] for syn in self._syncoutputs]:
            sync = [syn[1] for syn in self._syncoutputs if syn[0] == chan][0]
            sync_duration = 1000*self.channels[chan-1].sync_duration.get()
            sync_delay = 1000*self.channels[chan-1].sync_delay.get()
            self.write('syn {} {} {} {}'.format(sync, fg,
                                                sync_delay,
                                                sync_duration))

        typedict = {'SINE': 1, 'SQUARE': 2, 'RAMP': 3}

        typeval = typedict['RAMP']
        dutyval = 100
        # s -> ms
        periodval = ramptime*1e3
        repval = 1
        funmssg = 'fun {} {} {} {} {}'.format(fg,
                                              typeval, periodval,
                                              dutyval, repval)
        self.write(chanmssg)
        self.write(funmssg)

    def write(self, cmd: str) -> None:
        """
        QDac always returns something even from set commands, even when
        verbose mode is off, so we'll override write to take this out
        if you want to use this response, we put it in self._write_response
        (but only for the very last write call)

        In this method we expect to read one termination char per command. As
        commands are concatenated by `;` we count the number of concatenated
        commands as count(';') + 1 e.g. 'wav 1 1 1 0;fun 2 1 100 1 1' is two
        commands. Note that only the response of the last command will be
        available in `_write_response`

        """

        log.debug(f"Writing to instrument {self.name}: {cmd}")

        self.visa_handle.write(cmd)
        for _ in range(cmd.count(';')+1):
            self._write_response = self.visa_handle.read()

    def read(self) -> str:
        return self.visa_handle.read()

    def _wait_and_clear(self, delay: float = 0.5) -> None:
        time.sleep(delay)
        self.visa_handle.clear()

    def connect_message(self,
                        idn_part: str = "IDN",
                        being_time: Optional[float] = None) -> None:
        """
        Override of the standard Instrument class connect_message.
        Usually, the response to `*IDN?` is printed. Here, the
        software version is printed.
        """
        self.visa_handle.write('status')

        log.info('Connected to QDac on {}, {}'.format(self._address,
                                                      self.visa_handle.read()))

        # take care of the rest of the output
        for _ in range(self._output_n_lines):
            self.visa_handle.read()

    def _get_firmware_version(self) -> float:
        self.write('status')
        FW_str = self._write_response
        FW_version = float(FW_str.replace('Software Version: ', ''))
        for _ in range(self._output_n_lines):
            self.read()
        return FW_version

    def print_overview(self, update_currents: bool = False) -> None:
        """
        Pretty-prints the status of the QDac
        """

        self._update_cache(readcurrents=update_currents)

        paramstoget = [['i', 'v'], ['irange', 'vrange']]
        printdict = {'i': 'Current', 'v': 'Voltage', 'vrange': 'Voltage range',
                     'irange': 'Current range'}

        returnmap = {'vrange': {1: '-1 V to 1 V', 0: '-10 V to 10 V'},
                     'irange': {0: '0 to 1 muA', 1: '0 to 100 muA'}}

        # Print the channels
        for ii in range(self.num_chans):
            line = f"Channel {ii+1} \n"
            line += "    "
            for pp in paramstoget[0]:
                param = getattr(self.channels[ii], pp)
                line += printdict[pp]
                line += f': {param.get_latest()}'
                line += f' ({param.unit})'
                line += '. '
            line += '\n    '
            for pp in paramstoget[1]:
                param = getattr(self.channels[ii], pp)
                line += printdict[pp]
                value = param.get_latest()
                line += f": {returnmap[pp][value]}"
                line += ". "
            print(line)



import pytest

from qcodes.instrument_drivers.rohde_schwarz.private.HMC804x import _RohdeSchwarzHMC804x
import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'RSHMC804x.yaml@sim')


@pytest.fixture(scope='module')
def HMC8041():
    hmc8041 = _RohdeSchwarzHMC804x('hmc8041',
                                   address='GPIB::1::INSTR',
                                   num_channels=1,
                                   visalib=visalib,
                                   terminator='\n')
    yield hmc8041

    hmc8041.close()


@pytest.fixture(scope='module')
def HMC8042():
    hmc8042 = _RohdeSchwarzHMC804x('hmc8042',
                                   address='GPIB::1::INSTR',
                                   num_channels=2,
                                   visalib=visalib,
                                   terminator='\n')
    yield hmc8042

    hmc8042.close()


@pytest.fixture(scope='module')
def HMC8043():
    hmc8043 = _RohdeSchwarzHMC804x('hmc8043',
                                   address='GPIB::1::INSTR',
                                   num_channels=3,
                                   visalib=visalib,
                                   terminator='\n')
    yield hmc8043

    hmc8043.close()


def test_init(HMC8041, HMC8042, HMC8043):

    for hmc in [HMC8041, HMC8042, HMC8043]:

        idn_dict = hmc.IDN()

        assert idn_dict['vendor'] == 'QCoDeS'



"""
This module defines a :class:`logging.LoggerAdapter` and
:class:`logging.Filter`. They are used to enable the capturing of output from
specific
instruments.
"""

from contextlib import contextmanager
import logging
from typing import Optional, Sequence, Union, TYPE_CHECKING, Any, Tuple, \
    MutableMapping, Iterator
import collections.abc
from .logger import get_console_handler, LevelType, handler_level
if TYPE_CHECKING:
    import qcodes.instrument.InstrumentBase as InstrumentBase  # noqa: F401 pylint: disable=unused-import


class InstrumentLoggerAdapter(logging.LoggerAdapter):
    """
    In the Python logging module adapters are used to add context information
    to logging. The :class:`logging.LoggerAdapter` has the same methods as the
    :class:`logging.Logger` and can thus be used as such.

    Here it is used to add the instruments full name to the log records so that
    they can be filtered (using the :class:`InstrumentFilter`) by instrument
    instance.

    The context data gets stored in the `extra` dictionary as a property of the
    Adapter. It is filled by the ``__init__`` method::

        >>> LoggerAdapter(log, {'instrument': self.full_name})

    """
    def process(self, msg: str, kwargs: MutableMapping[str, Any]) -> \
            Tuple[str, MutableMapping[str, Any]]:
        """
        Returns the message and the kwargs for the handlers.
        """
        kwargs['extra'] = self.extra
        inst = self.extra['instrument']
        return f"[{inst.full_name}({type(inst).__name__})] {msg}", kwargs


class InstrumentFilter(logging.Filter):
    """
    Filter to filter out records that originate from the given instruments.
    Records created through the :class:`InstrumentLoggerAdapter` have additional
    properties as specified in the `extra` dictionary which is a property of
    the adapter.

    Here the ``instrument`` property gets used to reject records that don't
    originate from the list of instruments that has been passed to the
    ``__init__`` method.
    """
    def __init__(self, instruments: Union['InstrumentBase',
                                          Sequence['InstrumentBase']]):
        # This local import is necessary to avoid a circular import dependency.
        # The alternative is to merge this module with the instrument.base,
        # which is also not favorable.
        super().__init__()
        if not isinstance(instruments, collections.abc.Sequence):
            instrument_seq: Sequence['InstrumentBase'] = (instruments,)
        else:
            instrument_seq = instruments
        self.instrument_set = set(instrument_seq)

    def filter(self, record: logging.LogRecord) -> bool:
        inst = getattr(record, 'instrument', False)
        if inst is False:
            return False
        return not self.instrument_set.isdisjoint(inst.ancestors)


def get_instrument_logger(instrument_instance: 'InstrumentBase',
                          logger_name: Optional[str] = None
                          ) -> InstrumentLoggerAdapter:
    """
    Returns an :class:`InstrumentLoggerAdapter` that can be used to log
    messages
    including ``instrument_instance`` as  an additional context.

    The :class:`logging.LoggerAdapter` object can be used as any logger.

    Args:
        instrument_instance: The instrument instance to be added to the context
            of the log record.
        logger_name: Name of the logger to which the records will be passed.
            If `None`, defaults to the root logger.

    Returns:
        :class:`logging.LoggerAdapter` instance, that can be used for instrument
        specific logging.
    """
    logger_name = logger_name or ''
    return InstrumentLoggerAdapter(logging.getLogger(logger_name),
                                   {'instrument': instrument_instance})


@contextmanager
def filter_instrument(instrument: Union['InstrumentBase',
                                        Sequence['InstrumentBase']],
                      handler: Optional[
                          Union[logging.Handler,
                                Sequence[logging.Handler]]] = None,
                      level: Optional[LevelType] = None) -> Iterator[None]:
    """
    Context manager that adds a filter that only enables the log messages of
    the supplied instruments to pass.

    Example:
        >>> h1, h2 = logger.get_console_handler(), logger.get_file_handler()
        >>> with logger.filter_instruments((qdac, dmm2), handler=[h1, h2]):
        >>>     qdac.ch01(1)  # logged
        >>>     v1 = dmm2.v() # logged
        >>>     v2 = keithley.v()  # not logged

    Args:
        instrument: The instrument or sequence of instruments to enable
            messages from.
        level: Level to set the handlers to.
        handler: Single or sequence of handlers to change.
    """
    handlers: Sequence[logging.Handler]
    if handler is None:
        myhandler = get_console_handler()
        if myhandler is None:
            raise RuntimeError("Trying to filter instrument but no handler "
                               "defined. Did you forget to call "
                               "`start_logger` before?")
        handlers = (myhandler,)
    elif not isinstance(handler, collections.abc.Sequence):
        handlers = (handler,)
    else:
        handlers = handler

    instrument_filter = InstrumentFilter(instrument)
    for h in handlers:
        h.addFilter(instrument_filter)
    try:
        if level is not None:
            with handler_level(level, handlers):
                yield
        else:
            yield
    finally:
        for h in handlers:
            h.removeFilter(instrument_filter)



from typing import Any

from . import N52xx

#  This is not the same class of Keysight devices but seems to work well...

class P9374A(N52xx.PNAxBase):
    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(name, address,
                         min_freq=300e6, max_freq=20e9,
                         min_power=-43, max_power=20,
                         nports=2,
                         **kwargs)

        options = self.get_options()



import ctypes

import pytest

from qcodes.instrument_drivers.AlazarTech.ATS import Buffer


def test_buffer_is_allocated_when_initiated():
    b = Buffer(ctypes.c_uint8, 25)
    assert b._allocated is True


@pytest.mark.parametrize('ctype', (ctypes.c_uint8, ctypes.c_uint16,
                                   ctypes.c_uint32, ctypes.c_int32,
                                   ctypes.c_float))
def test_supported_ctypes_for_sample(ctype):
    Buffer(ctype, 8)



from typing import Any, Mapping, Optional, Sequence

from qcodes.instrument.delegate.delegate_instrument import DelegateInstrument
from qcodes.station import Station


class DelegateChannelInstrument(DelegateInstrument):
    """
    Delegate instrument that auto generates delegate parameters for a given
    ChannelList.

    Example usage in instrument YAML:

    .. code-block:: yaml

        switch:
            type: qcodes.instrument.delegate.DelegateChannelInstrument
            init:
                channels: dac.channels
                parameters:
                    state:
                    - dac_output
                    - smc
                    - gnd
                    - bus

    The above will create a new instrument called ``switch`` that generates a
    method for a delegate parameter:

    .. code-block:: python

        switch.state()

    that returns a named tuple:

    .. code-block:: python

        state(dac_output=..., smc=..., gnd=..., bus=...)

    where the values of each of the tuple items are delegated to the
    instrument parameters:

    .. code-block:: python

        dac.dac_output()
        dac.smc()
        dac.gnd()
        dac.bus()

    Args:
        name: Instrument name
        station: Station with real instruments to connect to
        channels: Path to channels, e.g. my_instrument.channels
        parameters: A mapping from name of a delegate parameter to the sequence
            of endpoint parameters it connects  to. These are auto-generated per
            channel.
        initial_values: Default values to set on instrument load. Defaults
            to None.
        set_initial_values_on_load: Flag to set defaults on load. Defaults
            to False.
    """
    def __init__(
        self,
        name: str,
        station: Station,
        channels: str,
        parameters: Mapping[str, Sequence[str]],
        initial_values: Optional[Mapping[str, Any]] = None,
        set_initial_values_on_load: bool = False,
        **kwargs: Any):
        _channels = self.parse_instrument_path(parent=station, path=channels)
        _parameters = {}
        for channel in _channels:
            ins = channel.root_instrument.name
            chan_no = str(channel.channel_number()).zfill(2)
            for alias, paths in parameters.items():
                _paths = [
                    f"{ins}.ch{chan_no}.{path}" for path in paths
                ]
                _parameters[f"{alias}{chan_no}"] = _paths

        super().__init__(
            name=name,
            station=station,
            parameters=_parameters,
            initial_values=initial_values,
            set_initial_values_on_load=set_initial_values_on_load,
            **kwargs
        )



"""
The module provides useful utility objects for the Alazar driver class.
The purpose of these is more to provide relevant functionality and less to
improve the code structure (which is more the purpose of the
:mod:`.AlazarTech.helpers` module).
"""

from typing import TYPE_CHECKING, Any, cast

from qcodes.instrument.parameter import Parameter, ParamRawDataType

if TYPE_CHECKING:
    from .ATS import AlazarTech_ATS


class TraceParameter(Parameter):
    """
    A parameter that keeps track of if its value has been synced to
    the ``Instrument``. To achieve that, this parameter sets
    the ``_parameters_synced`` attribute of the ``Instrument`` to ``False``
    when the value of the parameter is changed.

    This parameter is useful for the Alazar card driver. Syncing parameters to
    an Alazar card is relatively slow, hence it makes sense to first set the
    values of the parameters, and then "synchronize them to the card".
    """
    def __init__(self, *args: Any, **kwargs: Any):
        self._synced_to_card = False
        super().__init__(*args, **kwargs)

    def _set_updated(self) -> None:
        self._synced_to_card = True

    @property
    def synced_to_card(self) -> bool:
        """True if the parameter value has been synced to the instrument"""
        return self._synced_to_card

    def set_raw(self, value: ParamRawDataType) -> None:
        instrument = cast("AlazarTech_ATS", self.instrument)
        instrument._parameters_synced = False
        self._synced_to_card = False



from IPython.core.magic import Magics, magics_class, line_cell_magic

@magics_class
class QCoDeSMagic(Magics):
    """Magics related to code management (loading, saving, editing, ...)."""

    def __init__(self, *args, **kwargs):
        """
        Setup Magic. All args and kwargs are passed to super class.
        """
        self._knowntemps = set()
        super().__init__(*args, **kwargs)

    @line_cell_magic
    def measurement(self, line, cell=None):
        """
        Create ``qcodes.Loop`` measurement mimicking Python ``for`` syntax via
        iPython magic.

        Upon execution of a notebook cell, the code is transformed from the
        for loop structure to a QCoDeS Loop before being executed.
        Can be run by having ``%%measurement`` in the first line of a cell,
        followed by the measurement name (see below for an example).

        The for loop syntax differs slightly from a Python ``for`` loop,
        as it uses ``for {iterable}`` instead of ``for {element} in {iterable}``.
        The reason is that ``{element}`` cannot be accessed (yet) in QCoDeS loops.

        Comments (#) are ignored in the loop.
        Any code after the loop will also be run, if separated by a blank
        line from the loop.

        The Loop object is by default stored in a variable named ``loop``,
        and the dataset in ``data``, and these can be overridden using options.
        Must be run in a Jupyter Notebook.
        Delays can be provided in a loop by adding ``-d {delay}`` after ``for``.

        The following options can be passed along with the measurement name
        (e.g. ``%%measurement -px -d data_name {measurement_name})``::

            -p : print transformed code
            -x : Do not execute code
            -d <dataset_name> : Use custom name for dataset
            -l <loop_name> : Use custom name for Loop

        An example for a loop cell is as follows::

            %%measurement {-options} {measurement_name}
            for {sweep_vals}:
                {measure_parameter1}
                {measure_parameter2}
                for -d 1 {sweep_vals2}:
                    {measure_parameter3}

            ...

        which will be internally transformed to::

            import qcodes
            loop = qcodes.Loop({sweep_vals}).each(
                {measure_parameter1},
                {measure_parameter2},
                qcodes.Loop({sweep_vals2}, delay=1).each(
                    {measure_parameter3}))
            data = loop.get_data_set(name={measurement_name})

            ...

        An explicit example of the line ``for {sweep_vals}:`` could be
        ``for sweep_parameter.sweep(0, 42, step=1):``

        """

        if cell is None:
            # No loop provided, print documentation
            print(self.measurement.__doc__)
            return

        # Parse line, get measurement name and any possible options
        options, msmt_name = self.parse_options(line, 'pd:l:x')
        data_name = options.get('d', 'data')
        loop_name = options.get('l', 'loop')

        lines = cell.splitlines()
        assert lines[0][:3] == 'for', "Measurement must start with for loop"

        contents = f'import qcodes\n{loop_name} = '
        previous_level = 0
        for k, line in enumerate(lines):
            line, level = line.lstrip(), int((len(line)-len(line.lstrip())) / 4)

            if not line:
                # Empty line, end of loop
                break
            elif line[0] == '#':
                # Ignore comment
                continue
            else:
                line_representation = ' ' * level * 4
                if level < previous_level:
                    # Exiting inner loop, close bracket
                    line_representation += '),' * (previous_level - level)
                    line_representation += '\n' + ' ' * level * 4

                if line[:3] == 'for':
                    # New loop
                    for_opts, for_code = self.parse_options(line[4:-1], 'd:')
                    if 'd' in for_opts:
                        # Delay option provided
                        line_representation += ('qcodes.Loop({}, '
                                                'delay={}).each(\n'
                                                ''.format(for_code,
                                                          for_opts["d"]))
                    else:
                        line_representation += ('qcodes.Loop({}).each(\n'
                                                ''.format(for_code))
                else:
                    # Action in current loop
                    line_representation += f'{line},\n'
                contents += line_representation

                # Remember level for next iteration (might exit inner loop)
                previous_level = level

        # Add closing brackets for any remaining loops
        contents += ')' * previous_level + '\n'
        # Add dataset
        contents += "{} = {}.get_data_set(name='{}')".format(data_name,
                                                             loop_name,
                                                             msmt_name)

        for line in lines[k+1:]:
            contents += '\n' + line

        if 'p' in options:
            print(contents)

        if 'x' not in options:
            # Execute contents
            self.shell.run_cell(contents, store_history=True, silent=True)


def register_magic_class(cls=QCoDeSMagic, magic_commands=True):
    """
    Registers a iPython magic class.

    Args:
        cls: Magic class to register.
        magic_commands (List): List of magic commands within the class to
            register. If not specified, all magic commands are registered.

    """

    ip = get_ipython()
    if ip is None:
        raise RuntimeError('No iPython shell found')
    else:
        if magic_commands is not True:
            # filter out any magic commands that are not in magic_commands
            cls.magics = {line_cell: {key: val for key, val in magics.items()
                                      if key in magic_commands}
                          for line_cell, magics in cls.magics.items()}
        ip.magics_manager.register(cls)



import pytest

from qcodes.instrument_drivers.rohde_schwarz.RTO1000 import RTO1000
import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'RTO_1000.yaml@sim')


@pytest.fixture(scope='function')
def driver():
    rto_sim = RTO1000('rto_sim',
                      address='GPIB::1::INSTR',
                      visalib=visalib,
                      model='RTO1044'
                      )
    yield rto_sim

    rto_sim.close()


def test_init(driver):

    idn_dict = driver.IDN()

    assert idn_dict['vendor'] == 'QCoDeS'


def test_trigger_source_level(driver):
    assert driver.trigger_source() == 'CH1'
    assert driver.trigger_level() == 0
    driver.trigger_level(1.0)
    assert driver.trigger_level() == 1
    driver.trigger_level(0)



from typing import Any

from qcodes import VisaInstrument, validators as vals
import numpy as np
from qcodes.utils.validators import Numbers, numbertypes


class E8267(VisaInstrument):
    """
    This is the code for Agilent E8267 Signal Generator.
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address,  terminator='\n', **kwargs)
        # general commands
        self.add_parameter(name='frequency',
                           label='Frequency',
                           unit='Hz',
                           get_cmd='FREQ?',
                           set_cmd='FREQ {}',
                           get_parser=float,
                           vals=Numbers(min_value=100e3,
                                        max_value=40e9))
        self.add_parameter(name='freq_offset',
                           label='Frequency offset',
                           unit='Hz',
                           get_cmd='FREQ:OFFS?',
                           set_cmd='FREQ:OFFS {}',
                           get_parser=float,
                           vals=Numbers(min_value=-200e9,
                                        max_value=200e9))
        self.add_parameter('freq_mode',
                           label='Frequency mode',
                           set_cmd='FREQ:MODE {}',
                           get_cmd='FREQ:MODE?',
                           vals=vals.Enum('FIX', 'CW', 'SWE', 'LIST'))
        self.add_parameter('pulse_width',
                           label='Pulse width',
                           unit='ns',
                           set_cmd='PULM:INT:PWID {}',
                           get_cmd='PULM:INT:PWID?',
                           vals=Numbers(min_value=10e-9,
                                        max_value=20e-9))
        self.add_parameter(name='phase',
                           label='Phase',
                           unit='deg',
                           get_cmd='PHAS?',
                           set_cmd='PHAS {}',
                           get_parser=self.rad_to_deg,
                           set_parser=self.deg_to_rad,
                           vals=Numbers(min_value=-180,
                                        max_value=179))
        self.add_parameter(name='power',
                           label='Power',
                           unit='dBm',
                           get_cmd='POW?',
                           set_cmd='POW {}',
                           get_parser=float,
                           vals=Numbers(min_value=-135,
                                        max_value=25))
        self.add_parameter(name='power_offset',
                           label='Power offset',
                           unit='dBm',
                           get_cmd='POW:OFFS?',
                           set_cmd='POW:OFFS {}',
                           get_parser=float,
                           vals=Numbers(min_value=-200,
                                        max_value=200))
        self.add_parameter(name='output_rf',
                           get_cmd='OUTP?',
                           set_cmd='OUTP {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        self.add_parameter(name='modulation_rf',
                           get_cmd='OUTP:MOD?',
                           set_cmd='OUTP:MOD {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        # reset values after each reconnect
        self.power(0)
        self.power_offset(0)
        self.connect_message()
        self.add_function('reset', call_cmd='*RST')

    # functions to convert between rad and deg
    @staticmethod
    def deg_to_rad(angle_deg: numbertypes) -> "np.floating[Any]":
        return np.deg2rad(float(angle_deg))

    @staticmethod
    def rad_to_deg(angle_rad: numbertypes) -> "np.floating[Any]":
        return np.rad2deg(float(angle_rad))



import pytest

from qcodes.tests.instrument_mocks import DummyInstrument
from qcodes.instrument.parameter import ScaledParameter, ManualParameter


@pytest.fixture()
def instrument():
    instrument = DummyInstrument('dummy')
    try:
        target_name = 'target_parameter'
        target_label = 'Target Parameter'
        target_unit = 'V'
        instrument.add_parameter(
            target_name,
            label=target_label,
            unit=target_unit,
            initial_value=1.0,
            get_cmd=None,
            set_cmd=None
        )
        instrument.scaler = ScaledParameter(
            instrument.target_parameter,
            division=1
        )
        yield instrument
    finally:
        instrument.close()


def test_constructor(instrument):
    # Test the behaviour of the constructor

    # Require a wrapped parameter
    with pytest.raises(TypeError):
        ScaledParameter()

    # Require a scaling factor
    with pytest.raises(ValueError):
        ScaledParameter(instrument.target_parameter)

    # Require only one scaling factor
    with pytest.raises(ValueError):
        ScaledParameter(instrument.target_parameter, division=1, gain=1)


def test_namelabel(instrument):
    # Test handling of name and label

    # Test correct inheritance
    assert instrument.scaler.name == instrument.target_parameter.name + '_scaled'
    assert instrument.scaler.label == instrument.target_parameter.label + '_scaled'

    # Test correct name/label handling by the constructor
    scaled_name = 'scaled'
    scaled_label = "Scaled parameter"
    scaler2 = ScaledParameter(instrument.target_parameter, division=1,
                              name=scaled_name, label=scaled_label)
    assert scaler2.name == scaled_name
    assert scaler2.label == scaled_label


def test_unit(instrument):
    # Test handling of the units

    # Check if the unit is correctly inherited
    assert instrument.scaler.unit == 'V'

    # Check if we can change successfully the unit
    instrument.scaler.unit = 'A'
    assert instrument.scaler.unit == 'A'

    # Check if unit is correctly set in the constructor
    scaler2 = ScaledParameter(instrument.target_parameter,
                              name='scaled_value',
                              division=1,
                              unit='K')
    assert scaler2.unit == 'K'


def test_metadata(instrument):
    # Test the metadata

    test_gain = 3
    test_unit = 'V'
    instrument.scaler.gain = test_gain
    instrument.scaler.unit = test_unit

    # Check if relevant fields are present in the snapshot
    snap = instrument.scaler.snapshot()
    snap_keys = snap.keys()
    metadata_keys = snap['metadata'].keys()
    assert 'division' in snap_keys
    assert 'gain' in snap_keys
    assert 'role' in snap_keys
    assert 'unit' in snap_keys
    assert 'variable_multiplier' in metadata_keys
    assert 'wrapped_parameter' in metadata_keys
    assert 'wrapped_instrument' in metadata_keys

    # Check if the fields are correct
    assert snap['gain'] == test_gain
    assert snap['division'] == 1/test_gain
    assert snap['role'] == ScaledParameter.Role.GAIN
    assert snap['unit'] == test_unit
    assert snap['metadata']['variable_multiplier'] is False
    assert snap['metadata']['wrapped_parameter'] == instrument.target_parameter.name


def test_wrapped_parameter(instrument):
    # Test if the target parameter is correctly inherited
    assert instrument.scaler.wrapped_parameter == instrument.target_parameter


def test_divider(instrument):
    test_division = 10
    test_value = 5

    instrument.scaler.division = test_division
    instrument.scaler(test_value)
    assert instrument.scaler() == test_value
    assert instrument.target_parameter() == test_division * test_value
    assert instrument.scaler.gain == 1/test_division
    assert instrument.scaler.role == ScaledParameter.Role.DIVISION


def test_multiplier(instrument):
    test_multiplier = 10
    test_value = 5

    instrument.scaler.gain = test_multiplier
    instrument.scaler(test_value)
    assert instrument.scaler() == test_value
    assert instrument.target_parameter() == test_value / test_multiplier
    assert instrument.scaler.division == 1/test_multiplier
    assert instrument.scaler.role == ScaledParameter.Role.GAIN




def test_variable_gain(instrument):
    test_value = 5

    initial_gain = 2
    variable_gain_name = 'gain'
    gain = ManualParameter(name=variable_gain_name,
                           initial_value=initial_gain)
    instrument.scaler.gain = gain
    instrument.scaler(test_value)

    assert instrument.scaler() == test_value
    assert instrument.target_parameter() == test_value / initial_gain
    assert instrument.scaler.division == 1/initial_gain

    second_gain = 7
    gain(second_gain)
    # target value must change on scaler value change, not on gain/division
    assert instrument.target_parameter() == test_value / initial_gain
    instrument.scaler(test_value)
    assert instrument.target_parameter() == test_value / second_gain
    assert instrument.scaler.division == 1 / second_gain

    assert instrument.scaler.metadata['variable_multiplier'] == variable_gain_name



from datetime import datetime, timedelta
import time

import pytest

from qcodes.instrument.parameter import Parameter, _BaseParameter
from .conftest import BetterGettableParam


def test_get_latest():
    time_resolution = time.get_clock_info('time').resolution
    sleep_delta = 2 * time_resolution

    # Create a gettable parameter
    local_parameter = Parameter('test_param', set_cmd=None, get_cmd=None)
    before_set = datetime.now()
    time.sleep(sleep_delta)
    local_parameter.set(1)
    time.sleep(sleep_delta)
    after_set = datetime.now()

    # Check we return last set value, with the correct timestamp
    assert local_parameter.get_latest() == 1
    assert before_set < local_parameter.get_latest.get_timestamp() < after_set

    # Check that updating the value updates the timestamp
    time.sleep(sleep_delta)
    local_parameter.set(2)
    assert local_parameter.get_latest() == 2
    assert local_parameter.get_latest.get_timestamp() > after_set


def test_get_latest_raw_value():
    # To have a simple distinction between raw value and value of the
    # parameter lets create a parameter with an offset
    p = Parameter('p', set_cmd=None, get_cmd=None, offset=42)
    assert p.get_latest.get_timestamp() is None

    # Initially, the parameter's raw value is None
    assert p.get_latest.get_raw_value() is None

    # After setting the parameter to some value, the
    # ``.get_latest.get_raw_value()`` call should return the new raw value
    # of the parameter
    p(3)
    assert p.get_latest.get_timestamp() is not None
    assert p.get_latest.get() == 3
    assert p.get_latest() == 3
    assert p.get_latest.get_raw_value() == 3 + 42


def test_get_latest_unknown():
    """
    Test that get latest on a parameter that has not been acquired will
    trigger a get
    """
    value = 1
    local_parameter = BetterGettableParam('test_param', set_cmd=None,
                                          get_cmd=None)
    # fake a parameter that has a value but never been get/set to mock
    # an instrument.
    local_parameter.cache._value = value
    local_parameter.cache._raw_value = value
    assert local_parameter.get_latest.get_timestamp() is None
    before_get = datetime.now()
    assert local_parameter._get_count == 0
    assert local_parameter.get_latest() == value
    assert local_parameter._get_count == 1
    # calling get_latest above will call get since TS is None
    # and the TS will therefore no longer be None
    assert local_parameter.get_latest.get_timestamp() is not None
    assert local_parameter.get_latest.get_timestamp() >= before_get
    # calling get_latest now will not trigger get
    assert local_parameter.get_latest() == value
    assert local_parameter._get_count == 1


def test_get_latest_known():
    """
    Test that get latest on a parameter that has a known value will not
    trigger a get
    """
    value = 1
    local_parameter = BetterGettableParam('test_param', set_cmd=None,
                                          get_cmd=None)
    # fake a parameter that has a value acquired 10 sec ago
    start = datetime.now()
    set_time = start - timedelta(seconds=10)
    local_parameter.cache._update_with(
        value=value, raw_value=value, timestamp=set_time)
    assert local_parameter._get_count == 0
    assert local_parameter.get_latest.get_timestamp() == set_time
    assert local_parameter.get_latest() == value
    # calling get_latest above will not call get since TS is set and
    # max_val_age is not
    assert local_parameter._get_count == 0
    assert local_parameter.get_latest.get_timestamp() == set_time


def test_get_latest_no_get():
    """
    Test that get_latest on a parameter that does not have get is handled
    correctly.
    """
    local_parameter = Parameter('test_param', set_cmd=None, get_cmd=False)
    # The parameter does not have a get method.
    with pytest.raises(AttributeError):
        local_parameter.get()
    # get_latest will fail as get cannot be called and no cache
    # is available
    with pytest.raises(RuntimeError):
        local_parameter.get_latest()
    value = 1
    local_parameter.set(value)
    assert local_parameter.get_latest() == value

    local_parameter2 = Parameter('test_param2', set_cmd=None,
                                 get_cmd=False, initial_value=value)
    with pytest.raises(AttributeError):
        local_parameter2.get()
    assert local_parameter2.get_latest() == value


def test_max_val_age():
    value = 1
    start = datetime.now()
    local_parameter = BetterGettableParam('test_param',
                                          set_cmd=None,
                                          max_val_age=1,
                                          initial_value=value)
    assert local_parameter.cache.max_val_age == 1
    assert local_parameter._get_count == 0
    assert local_parameter.get_latest() == value
    assert local_parameter._get_count == 0
    # now fake the time stamp so get should be triggered
    set_time = start - timedelta(seconds=10)
    local_parameter.cache._update_with(
        value=value, raw_value=value, timestamp=set_time)
    # now that ts < max_val_age calling get_latest should update the time
    assert local_parameter.get_latest.get_timestamp() == set_time
    assert local_parameter.get_latest() == value
    assert local_parameter._get_count == 1
    assert local_parameter.get_latest.get_timestamp() >= start


def test_no_get_max_val_age():
    """
    Test that get_latest on a parameter with max_val_age set and
    no get cmd raises correctly.
    """
    value = 1
    with pytest.raises(SyntaxError):
        _ = Parameter('test_param', set_cmd=None,
                      get_cmd=False,
                      max_val_age=1, initial_value=value)

    # _BaseParameter does not have this check on creation time since get_cmd
    # could be added in a subclass. Here we create a subclass that does add a
    # get command and also does not implement the check for max_val_age
    class LocalParameter(_BaseParameter):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.set_raw = lambda x: x
            self.set = self._wrap_set(self.set_raw)

    localparameter = LocalParameter('test_param',
                                    None,
                                    max_val_age=1)
    with pytest.raises(RuntimeError):
        localparameter.get_latest()



from typing import Dict, ClassVar, Any

from qcodes.instrument_drivers.Lakeshore.lakeshore_base import (
    LakeshoreBase, BaseOutput, BaseSensorChannel)
from qcodes.instrument.group_parameter import GroupParameter, Group
import qcodes.utils.validators as vals


# There are 16 sensors channels (a.k.a. measurement inputs) in Model 372
_n_channels = 16


class Output_372(BaseOutput):
    """Class for control outputs (heaters) of model 372"""

    MODES: ClassVar[Dict[str, int]] = {
        'off': 0,
        'monitor_out': 1,
        'open_loop': 2,
        'zone': 3,
        'still': 4,
        'closed_loop': 5,
        'warm_up': 6}
    POLARITIES: ClassVar[Dict[str, int]] = {
        'unipolar': 0,
        'bipolar': 1}
    RANGES: ClassVar[Dict[str, int]] = {
        'off': 0,
        '31.6μA': 1,
        '100μA': 2,
        '316μA': 3,
        '1mA': 4,
        '3.16mA': 5,
        '10mA': 6,
        '31.6mA': 7,
        '100mA': 8}

    _input_channel_parameter_kwargs: ClassVar[Dict[str, Any]] = {
        'get_parser': int,
        'vals': vals.Numbers(1, _n_channels)}

    def __init__(
            self,
            parent: "Model_372",
            output_name: str,
            output_index: int
    ) -> None:
        super().__init__(parent, output_name, output_index, has_pid=True)

        # Add more parameters for OUTMODE command
        # and redefine the corresponding group
        self.add_parameter('polarity',
                           label='Output polarity',
                           docstring='Specifies output polarity (not '
                                     'applicable to warm-up heater)',
                           val_mapping=self.POLARITIES,
                           parameter_class=GroupParameter)
        self.add_parameter('use_filter',
                           label='Use filter for readings',
                           docstring='Specifies controlling on unfiltered or '
                                     'filtered readings',
                           val_mapping={True: 1, False: 0},
                           parameter_class=GroupParameter)
        self.add_parameter('delay',
                           label='Delay',
                           unit='s',
                           docstring='Delay in seconds for setpoint change '
                                     'during Autoscanning',
                           vals=vals.Ints(0, 255),
                           get_parser=int,
                           parameter_class=GroupParameter)
        self.output_group = Group([self.mode, self.input_channel,
                                   self.powerup_enable, self.polarity,
                                   self.use_filter, self.delay],
                                  set_cmd=f'OUTMODE {output_index}, {{mode}}, '
                                          f'{{input_channel}}, '
                                          f'{{powerup_enable}}, {{polarity}}, '
                                          f'{{use_filter}}, {{delay}}',
                                  get_cmd=f'OUTMODE? {output_index}')

        self.P.vals = vals.Numbers(0.0, 1000)
        self.I.vals = vals.Numbers(0.0, 10000)
        self.D.vals = vals.Numbers(0, 2500)


class Model_372_Channel(BaseSensorChannel):
    SENSOR_STATUSES = {0: 'OK',
                       1: 'CS OVL',
                       2: 'VCM OVL',
                       4: 'VMIX OVL',
                       8: 'VDIF OVL',
                       16: 'R. OVER',
                       32: 'R. UNDER',
                       64: 'T. OVER',
                       128: 'T. UNDER'}

    def __init__(
            self,
            parent: "Model_372",
            name: str,
            channel: str
    ):
        super().__init__(parent, name, channel)

        # Parameters related to Input Channel Parameter Command (INSET)
        self.add_parameter('enabled',
                           label='Enabled',
                           docstring='Specifies whether the input/channel is '
                                     'enabled or disabled. At least one '
                                     'measurement input channel must be '
                                     'enabled. If all are configured to '
                                     'disabled, channel 1 will change to '
                                     'enabled.',
                           val_mapping={True: 1, False: 0},
                           parameter_class=GroupParameter)
        self.add_parameter('dwell',
                           label='Dwell',
                           docstring='Specifies a value for the autoscanning '
                                     'dwell time.',
                           unit='s',
                           get_parser=int,
                           vals=vals.Numbers(1, 200),
                           parameter_class=GroupParameter)
        self.add_parameter('pause',
                           label='Change pause time',
                           docstring='Specifies a value for '
                                     'the change pause time',
                           unit='s',
                           get_parser=int,
                           vals=vals.Numbers(3, 200),
                           parameter_class=GroupParameter)
        self.add_parameter('curve_number',
                           label='Curve',
                           docstring='Specifies which curve the channel uses: '
                                     '0 = no curve, 1 to 59 = standard/user '
                                     'curves. Do not change this parameter '
                                     'unless you know what you are doing.',
                           get_parser=int,
                           vals=vals.Numbers(0, 59),
                           parameter_class=GroupParameter)
        self.add_parameter('temperature_coefficient',
                           label='Change pause time',
                           docstring='Sets the temperature coefficient that '
                                     'will be used for temperature control if '
                                     'no curve is selected (negative or '
                                     'positive). Do not change this parameter '
                                     'unless you know what you are doing.',
                           val_mapping={'negative': 1, 'positive': 2},
                           parameter_class=GroupParameter)
        self.output_group = Group([self.enabled, self.dwell, self.pause,
                                   self.curve_number,
                                   self.temperature_coefficient],
                                  set_cmd=f'INSET {self._channel}, '
                                          f'{{enabled}}, {{dwell}}, {{pause}}, '
                                          f'{{curve_number}}, '
                                          f'{{temperature_coefficient}}',
                                  get_cmd=f'INSET? {self._channel}')

        # Parameters related to Input Setup Command (INTYPE)
        self.add_parameter('excitation_mode',
                           label='Excitation mode',
                           docstring='Specifies excitation mode',
                           val_mapping={'voltage': 0, 'current': 1},
                           parameter_class=GroupParameter)
        # The allowed values for this parameter change based on the value of
        # the 'excitation_mode' parameter. Moreover, there is a table in the
        # manual that assigns the numbers to particular voltage/current ranges.
        # Once this parameter is heavily used, it can be implemented properly
        # (i.e. using val_mapping, and that val_mapping is updated based on the
        # value of 'excitation_mode'). At the moment, this parameter is added
        # only because it is a part of a group.
        self.add_parameter('excitation_range_number',
                           label='Excitation range number',
                           docstring='Specifies excitation range number '
                                     '(1-12 for voltage excitation, 1-22 for '
                                     'current excitation); refer to the manual '
                                     'for the table of ranges',
                           get_parser=int,
                           vals=vals.Numbers(1, 22),
                           parameter_class=GroupParameter)
        self.add_parameter('auto_range',
                           label='Auto range',
                           docstring='Specifies auto range setting',
                           val_mapping={'off': 0, 'current': 1},
                           parameter_class=GroupParameter)
        self.add_parameter('range',
                           label='Range',
                           val_mapping={'2.0 mOhm': 1,
                                        '6.32 mOhm': 2,
                                        '20.0 mOhm': 3,
                                        '63.2 mOhm': 4,
                                        '200 mOhm': 5,
                                        '632 mOhm': 6,
                                        '2.00 Ohm': 7,
                                        '6.32 Ohm': 8,
                                        '20.0 Ohm': 9,
                                        '63.2 Ohm': 10,
                                        '200 Ohm': 11,
                                        '632 Ohm': 12,
                                        '2.00 kOhm': 13,
                                        '6.32 kOhm': 14,
                                        '20.0 kOhm': 15,
                                        '63.2 kOhm': 16,
                                        '200 kOhm': 17,
                                        '632 kOhm': 18,
                                        '2.0 MOhm': 19,
                                        '6.32 MOhm': 20,
                                        '20.0 MOhm': 21,
                                        '63.2 MOhm': 22},
                           parameter_class=GroupParameter)
        self.add_parameter('current_source_shunted',
                           label='Current source shunt',
                           docstring='Current source either not shunted '
                                     '(excitation on), or shunted '
                                     '(excitation off)',
                           val_mapping={False: 0, True: 1},
                           parameter_class=GroupParameter)
        self.add_parameter('units',
                           label='Preferred units',
                           docstring='Specifies the preferred units parameter '
                                     'for sensor readings and for the control '
                                     'setpoint (kelvin or ohms)',
                           val_mapping={'kelvin': 1, 'ohms': 2},
                           parameter_class=GroupParameter)
        self.output_group = Group([self.excitation_mode,
                                   self.excitation_range_number,
                                   self.auto_range, self.range,
                                   self.current_source_shunted, self.units],
                                  set_cmd=f'INTYPE {self._channel}, '
                                          f'{{excitation_mode}}, '
                                          f'{{excitation_range_number}}, '
                                          f'{{auto_range}}, {{range}}, '
                                          f'{{current_source_shunted}}, '
                                          f'{{units}}',
                                  get_cmd=f'INTYPE? {self._channel}')


class Model_372(LakeshoreBase):
    """
    Lakeshore Model 372 Temperature Controller Driver

    Note that interaction with the control input (referred to as 'A' in the
    Computer Interface Operation section of the manual) is not implemented.
    """
    channel_name_command: Dict[str, str] = {f'ch{i:02}': str(i)
                                            for i in range(1, 1 + _n_channels)}
    input_channel_parameter_values_to_channel_name_on_instrument = {
        i: f'ch{i:02}' for i in range(1, 1 + _n_channels)
    }

    CHANNEL_CLASS = Model_372_Channel

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, **kwargs)

        heaters = {'sample_heater': 0, 'warmup_heater': 1, 'analog_heater': 2}
        for heater_name, heater_index in heaters.items():
            self.add_submodule(heater_name, Output_372(self, heater_name, heater_index))



import logging

import pytest
import hypothesis.strategies as hst
from hypothesis import given, settings
import numpy as np

from .conftest import MemoryParameter
from qcodes.utils.validators import Numbers
from qcodes.instrument.parameter import Parameter


def test_step_ramp(caplog):
    p = MemoryParameter(name='test_step')
    p(42)
    assert p.set_values == [42]
    p.step = 1

    assert p.get_ramp_values(44.5, 1) == [43, 44, 44.5]

    p(44.5)
    assert p.set_values == [42, 43, 44, 44.5]

    # Assert that stepping does not impact ``cache.set`` call, and that
    # the value that is passed to ``cache.set`` call does not get
    # propagated to parameter's ``set_cmd``
    p.cache.set(40)
    assert p.get_latest() == 40
    assert p.set_values == [42, 43, 44, 44.5]

    # Test error conditions
    with caplog.at_level(logging.WARNING):
        assert p.get_ramp_values("A", 1) == ["A"]
        assert len(caplog.records) == 1
        assert "cannot sweep test_step from 40 to 'A'" in str(caplog.records[0])
    with pytest.raises(RuntimeError):
        p.get_ramp_values((1, 2, 3), 1)


@given(scale=hst.integers(1, 100),
       value=hst.floats(min_value=1e-9, max_value=10))
def test_ramp_scaled(scale, value):
    start_point = 0.0
    p = MemoryParameter(name='p', scale=scale,
                        initial_value=start_point)
    assert p() == start_point
    # first set a step size
    p.step = 0.1
    # and a wait time
    p.inter_delay = 1e-9 # in seconds
    first_step = 1.0
    second_step = 10.0
    # do a step to start from a non zero starting point where
    # scale matters
    p.set(first_step)
    np.testing.assert_allclose(np.array([p.get()]),
                               np.array([first_step]))

    expected_raw_steps = np.linspace(start_point*scale, first_step*scale, 11)
    # getting the raw values that are actually send to the instrument.
    # these are scaled in the set_wrapper
    np.testing.assert_allclose(np.array(p.set_values), expected_raw_steps)
    assert p.raw_value == first_step*scale
    # then check the generated steps. They should not be scaled as the
    # scaling happens when setting them
    expected_steps = np.linspace(first_step+p.step,
                                 second_step,90)
    np.testing.assert_allclose(p.get_ramp_values(second_step, p.step),
                               expected_steps)
    p.set(10)
    np.testing.assert_allclose(np.array(p.set_values),
                               np.linspace(0.0*scale, 10*scale, 101))
    p.set(value)
    np.testing.assert_allclose(p.get(), value)
    assert p.raw_value == value * scale


@given(value=hst.floats(min_value=1e-9, max_value=10))
def test_ramp_parser(value):
    start_point = 0.0
    p = MemoryParameter(name='p',
                        set_parser=lambda x: -x,
                        get_parser=lambda x: -x,
                        initial_value=start_point)
    assert p() == start_point
    # first set a step size
    p.step = 0.1
    # and a wait time
    p.inter_delay = 1e-9 # in seconds
    first_step = 1.0
    second_step = 10.0
    # do a step to start from a non zero starting point where
    # scale matters
    p.set(first_step)
    assert p.get() == first_step
    assert p.raw_value == - first_step
    np.testing.assert_allclose(np.array([p.get()]),
                               np.array([first_step]))

    expected_raw_steps = np.linspace(-start_point, -first_step, 11)
    # getting the raw values that are actually send to the instrument.
    # these are parsed in the set_wrapper
    np.testing.assert_allclose(np.array(p.set_values), expected_raw_steps)
    assert p.raw_value == -first_step
    # then check the generated steps. They should not be parsed as the
    # scaling happens when setting them
    expected_steps = np.linspace((first_step+p.step),
                                 second_step,90)
    np.testing.assert_allclose(p.get_ramp_values(second_step, p.step),
                               expected_steps)
    p.set(second_step)
    np.testing.assert_allclose(np.array(p.set_values),
                               np.linspace(-start_point, -second_step, 101))
    p.set(value)
    np.testing.assert_allclose(p.get(), value)
    assert p.raw_value == - value


@given(scale=hst.integers(1, 100),
       value=hst.floats(min_value=1e-9, max_value=10))
@settings(deadline=None)
def test_ramp_parsed_scaled(scale, value):
    start_point = 0.0
    p = MemoryParameter(name='p',
                        scale=scale,
                        set_parser=lambda x: -x,
                        get_parser=lambda x: -x,
                        initial_value=start_point)
    assert p() == start_point
    # first set a step size
    p.step = 0.1
    # and a wait time
    p.inter_delay = 1e-9  # in seconds
    first_step = 1.0
    second_step = 10.0
    p.set(first_step)
    assert p.get() == first_step
    assert p.raw_value == - first_step * scale
    expected_raw_steps = np.linspace(-start_point*scale, -first_step*scale, 11)
    # getting the raw values that are actually send to the instrument.
    # these are parsed in the set_wrapper
    np.testing.assert_allclose(np.array(p.set_values), expected_raw_steps)
    assert p.raw_value == - scale * first_step
    expected_steps = np.linspace(first_step+p.step, second_step, 90)
    np.testing.assert_allclose(p.get_ramp_values(10, p.step),
                               expected_steps)
    p.set(second_step)
    np.testing.assert_allclose(np.array(p.set_values),
                               np.linspace(-start_point*scale,
                                           -second_step*scale, 101))
    p.set(value)
    np.testing.assert_allclose(p.get(), value)
    assert p.raw_value == -scale * value


def test_stepping_from_invalid_starting_point():

    the_value = -10

    def set_function(value):
        nonlocal the_value
        the_value = value

    def get_function():
        return the_value

    a = Parameter('test', set_cmd=set_function, get_cmd=get_function,
                  vals=Numbers(0, 100), step=5)
    # We start out by setting the parameter to an
    # invalid value. This is not possible using initial_value
    # as the validator will catch that but perhaps this may happen
    # if the instrument can return out of range values.
    assert a.get() == -10
    with pytest.raises(ValueError):
        # trying to set to 10 should raise even with 10 valid
        # as the steps demand that we first step to -5 which is not
        a.set(10)
    # afterwards the value should still be the same
    assert a.get() == -10



import os
import sys

from setuptools import setup

# we need to import versioneer from the current dir
# once versioneer is pep517/518 compliant this can be removed
sys.path.append(os.path.dirname(__file__))
import versioneer

sys.path.pop()


if __name__ == "__main__":
    setup(
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
    )



import logging
import re
import time
from typing import Any, Sequence, Union

import numpy as np
from pyvisa import VisaIOError, errors

from qcodes import ArrayParameter, ChannelList, InstrumentChannel, VisaInstrument
from qcodes.utils.validators import Bool, Enum, Ints, Numbers

logger = logging.getLogger()

class PNASweep(ArrayParameter):
    def __init__(self,
                 name: str,
                 instrument: 'PNABase',
                 **kwargs: Any) -> None:

        super().__init__(name,
                         instrument=instrument,
                         shape=(0,),
                         setpoints=((0,),),
                         **kwargs)

    @property  # type: ignore[override]
    def shape(self) -> Sequence[int]:  # type: ignore[override]
        if self.instrument is None:
            return (0,)
        return (self.instrument.root_instrument.points(),)

    @shape.setter
    def shape(self, val: Sequence[int]) -> None:
        pass

    @property  # type: ignore[override]
    def setpoints(self) -> Sequence[np.ndarray]:  # type: ignore[override]
        if self.instrument is None:
            raise RuntimeError("Cannot return setpoints if not attached "
                               "to instrument")
        start = self.instrument.root_instrument.start()
        stop = self.instrument.root_instrument.stop()
        return (np.linspace(start, stop, self.shape[0]),)

    @setpoints.setter
    def setpoints(self, val: Sequence[int]) -> None:
        pass


class FormattedSweep(PNASweep):
    """
    Mag will run a sweep, including averaging, before returning data.
    As such, wait time in a loop is not needed.
    """
    def __init__(self,
                 name: str,
                 instrument: 'PNABase',
                 sweep_format: str,
                 label: str,
                 unit: str,
                 memory: bool = False) -> None:
        super().__init__(name,
                         instrument=instrument,
                         label=label,
                         unit=unit,
                         setpoint_names=('frequency',),
                         setpoint_labels=('Frequency',),
                         setpoint_units=('Hz',)
                         )
        self.sweep_format = sweep_format
        self.memory = memory

    def get_raw(self) -> Sequence[float]:
        if self.instrument is None:
            raise RuntimeError("Cannot get data without instrument")
        root_instr = self.instrument.root_instrument
        # Check if we should run a new sweep
        if root_instr.auto_sweep():
            prev_mode = self.instrument.run_sweep()
        # Ask for data, setting the format to the requested form
        self.instrument.format(self.sweep_format)
        data = root_instr.visa_handle.query_binary_values('CALC:DATA? FDATA',
                                                          datatype='f',
                                                          is_big_endian=True)
        data = np.array(data)
        # Restore previous state if it was changed
        if root_instr.auto_sweep():
            root_instr.sweep_mode(prev_mode)

        return data


class PNAPort(InstrumentChannel):
    """
    Allow operations on individual PNA ports.
    Note: This can be expanded to include a large number of extra parameters...
    """

    def __init__(self,
                 parent: 'PNABase',
                 name: str,
                 port: int,
                 min_power: Union[int, float],
                 max_power: Union[int, float]) -> None:
        super().__init__(parent, name)

        self.port = int(port)
        if self.port < 1 or self.port > 4:
            raise ValueError("Port must be between 1 and 4.")

        pow_cmd = f"SOUR:POW{self.port}"
        self.add_parameter("source_power",
                           label="power",
                           unit="dBm",
                           get_cmd=f"{pow_cmd}?",
                           set_cmd=f"{pow_cmd} {{}}",
                           get_parser=float,
                           vals=Numbers(min_value=min_power,
                                        max_value=max_power))

    def _set_power_limits(self,
                          min_power: Union[int, float],
                          max_power: Union[int, float]) -> None:
        """
        Set port power limits
        """
        self.source_power.vals = Numbers(min_value=min_power,
                                         max_value=max_power)


class PNATrace(InstrumentChannel):
    """
    Allow operations on individual PNA traces.
    """

    def __init__(self,
                 parent: 'PNABase',
                 name: str,
                 trace_name: str,
                 trace_num: int) -> None:
        super().__init__(parent, name)
        self.trace_name = trace_name
        self.trace_num = trace_num

        # Name of parameter (i.e. S11, S21 ...)
        self.add_parameter('trace',
                           label='Trace',
                           get_cmd=self._Sparam,
                           set_cmd=self._set_Sparam)
        # Format
        # Note: Currently parameters that return complex values are not
        # supported as there isn't really a good way of saving them into the
        # dataset
        self.add_parameter('format',
                           label='Format',
                           get_cmd='CALC:FORM?',
                           set_cmd='CALC:FORM {}',
                           vals=Enum('MLIN', 'MLOG', 'PHAS',
                                     'UPH', 'IMAG', 'REAL'))

        # And a list of individual formats
        self.add_parameter('magnitude',
                           sweep_format='MLOG',
                           label='Magnitude',
                           unit='dB',
                           parameter_class=FormattedSweep)
        self.add_parameter('linear_magnitude',
                           sweep_format='MLIN',
                           label='Magnitude',
                           unit='ratio',
                           parameter_class=FormattedSweep)
        self.add_parameter('phase',
                           sweep_format='PHAS',
                           label='Phase',
                           unit='deg',
                           parameter_class=FormattedSweep)
        self.add_parameter('unwrapped_phase',
                           sweep_format='UPH',
                           label='Phase',
                           unit='deg',
                           parameter_class=FormattedSweep)
        self.add_parameter("group_delay",
                           sweep_format='GDEL',
                           label='Group Delay',
                           unit='s',
                           parameter_class=FormattedSweep)
        self.add_parameter('real',
                           sweep_format='REAL',
                           label='Real',
                           unit='LinMag',
                           parameter_class=FormattedSweep)
        self.add_parameter('imaginary',
                           sweep_format='IMAG',
                           label='Imaginary',
                           unit='LinMag',
                           parameter_class=FormattedSweep)

    def run_sweep(self) -> str:
        """
        Run a set of sweeps on the network analyzer.
        Note that this will run all traces on the current channel.
        """
        root_instr = self.root_instrument
        # Store previous mode
        prev_mode = root_instr.sweep_mode()
        # Take instrument out of continuous mode, and send triggers equal to
        # the number of averages
        if root_instr.averages_enabled():
            avg = root_instr.averages()
            root_instr.reset_averages()
            root_instr.group_trigger_count(avg)
            root_instr.sweep_mode('GRO')
        else:
            root_instr.sweep_mode('SING')

        # Once the sweep mode is in hold, we know we're done
        try:
            while root_instr.sweep_mode() != 'HOLD':
                time.sleep(0.1)
        except KeyboardInterrupt:
            # If the user aborts because (s)he is stuck in the infinite loop
            # mentioned above, provide a hint of what can be wrong.
            msg = "User abort detected. "
            source = root_instr.trigger_source()
            if source == "MAN":
                msg += "The trigger source is manual. Are you sure this is " \
                       "correct? Please set the correct source with the " \
                       "'trigger_source' parameter"
            elif source == "EXT":
                msg += "The trigger source is external. Is the trigger " \
                       "source functional?"
            logger.warning(msg)

        # Return previous mode, incase we want to restore this
        return prev_mode

    def write(self, cmd: str) -> None:
        """
        Select correct trace before querying
        """
        self.root_instrument.active_trace(self.trace_num)
        super().write(cmd)

    def ask(self, cmd: str) -> str:
        """
        Select correct trace before querying
        """
        self.root_instrument.active_trace(self.trace_num)
        return super().ask(cmd)

    def _Sparam(self) -> str:
        """
        Extrace S_parameter from returned PNA format
        """
        paramspec = self.root_instrument.get_trace_catalog()
        specs = paramspec.split(',')
        for spec_ind in range(len(specs)//2):
            name, param = specs[spec_ind*2:(spec_ind+1)*2]
            if name == self.trace_name:
                return param
        raise RuntimeError("Can't find selected trace on the PNA")

    def _set_Sparam(self, val: str) -> None:
        """
        Set an S-parameter, in the format S<a><b>, where a and b
        can range from 1-4
        """
        if not re.match("S[1-4][1-4]", val):
            raise ValueError("Invalid S parameter spec")
        self.write(f"CALC:PAR:MOD:EXT \"{val}\"")


class PNABase(VisaInstrument):
    """
    Base qcodes driver for Agilent/Keysight series PNAs
    http://na.support.keysight.com/pna/help/latest/Programming/GP-IB_Command_Finder/SCPI_Command_Tree.htm

    Note: Currently this driver only expects a single channel on the PNA. We
          can handle multiple traces, but using traces across multiple channels
          may have unexpected results.
    """

    def __init__(self,
                 name: str,
                 address: str,
                 # Set frequency ranges
                 min_freq: Union[int, float], max_freq: Union[int, float],
                 # Set power ranges
                 min_power: Union[int, float], max_power: Union[int, float],
                 nports: int, # Number of ports on the PNA
                 **kwargs: Any) -> None:
        super().__init__(name, address, terminator='\n', **kwargs)
        self.min_freq = min_freq
        self.max_freq = max_freq

        #Ports
        ports = ChannelList(self, "PNAPorts", PNAPort)
        for port_num in range(1, nports+1):
            port = PNAPort(self, f"port{port_num}", port_num,
                           min_power, max_power)
            ports.append(port)
            self.add_submodule(f"port{port_num}", port)
        ports.lock()
        self.add_submodule("ports", ports)

        # Drive power
        self.add_parameter('power',
                           label='Power',
                           get_cmd='SOUR:POW?',
                           get_parser=float,
                           set_cmd='SOUR:POW {:.2f}',
                           unit='dBm',
                           vals=Numbers(min_value=min_power,
                                        max_value=max_power))

        # IF bandwidth
        self.add_parameter('if_bandwidth',
                           label='IF Bandwidth',
                           get_cmd='SENS:BAND?',
                           get_parser=float,
                           set_cmd='SENS:BAND {:.2f}',
                           unit='Hz',
                           vals=Numbers(min_value=1, max_value=15e6))

        # Number of averages (also resets averages)
        self.add_parameter('averages_enabled',
                           label='Averages Enabled',
                           get_cmd="SENS:AVER?",
                           set_cmd="SENS:AVER {}",
                           val_mapping={True: '1', False: '0'})
        self.add_parameter('averages',
                           label='Averages',
                           get_cmd='SENS:AVER:COUN?',
                           get_parser=int,
                           set_cmd='SENS:AVER:COUN {:d}',
                           unit='',
                           vals=Numbers(min_value=1, max_value=65536))

        # Setting frequency range
        self.add_parameter('start',
                           label='Start Frequency',
                           get_cmd='SENS:FREQ:STAR?',
                           get_parser=float,
                           set_cmd='SENS:FREQ:STAR {}',
                           unit='Hz',
                           vals=Numbers(min_value=min_freq,
                                        max_value=max_freq))
        self.add_parameter('stop',
                           label='Stop Frequency',
                           get_cmd='SENS:FREQ:STOP?',
                           get_parser=float,
                           set_cmd='SENS:FREQ:STOP {}',
                           unit='Hz',
                           vals=Numbers(min_value=min_freq,
                                        max_value=max_freq))
        self.add_parameter('center',
                           label='Center Frequency',
                           get_cmd='SENS:FREQ:CENT?',
                           get_parser=float,
                           set_cmd='SENS:FREQ:CENT {}',
                           unit='Hz',
                           vals=Numbers(min_value=min_freq,
                                        max_value=max_freq))
        self.add_parameter('span',
                           label='Frequency Span',
                           get_cmd='SENS:FREQ:SPAN?',
                           get_parser=float,
                           set_cmd='SENS:FREQ:SPAN {}',
                           unit='Hz',
                           vals=Numbers(min_value=min_freq,
                                        max_value=max_freq))

        # Number of points in a sweep
        self.add_parameter('points',
                           label='Points',
                           get_cmd='SENS:SWE:POIN?',
                           get_parser=int,
                           set_cmd='SENS:SWE:POIN {}',
                           unit='',
                           vals=Numbers(min_value=1, max_value=100001))

        # Electrical delay
        self.add_parameter('electrical_delay',
                           label='Electrical Delay',
                           get_cmd='CALC:CORR:EDEL:TIME?',
                           get_parser=float,
                           set_cmd='CALC:CORR:EDEL:TIME {:.6e}',
                           unit='s',
                           vals=Numbers(min_value=0, max_value=100000))

        # Sweep Time
        self.add_parameter('sweep_time',
                           label='Time',
                           get_cmd='SENS:SWE:TIME?',
                           get_parser=float,
                           unit='s',
                           vals=Numbers(0, 1e6))
        # Sweep Mode
        self.add_parameter('sweep_mode',
                           label='Mode',
                           get_cmd='SENS:SWE:MODE?',
                           set_cmd='SENS:SWE:MODE {}',
                           vals=Enum("HOLD", "CONT", "GRO", "SING"))
        # Group trigger count
        self.add_parameter('group_trigger_count',
                           get_cmd="SENS:SWE:GRO:COUN?",
                           get_parser=int,
                           set_cmd="SENS:SWE:GRO:COUN {}",
                           vals=Ints(1, 2000000))
        # Trigger Source
        self.add_parameter('trigger_source',
                           get_cmd="TRIG:SOUR?",
                           set_cmd="TRIG:SOUR {}",
                           vals=Enum("EXT", "IMM", "MAN"))

        # Traces
        self.add_parameter('active_trace',
                           label='Active Trace',
                           get_cmd="CALC:PAR:MNUM?",
                           get_parser=int,
                           set_cmd="CALC:PAR:MNUM {}",
                           vals=Numbers(min_value=1, max_value=24))
        # Note: Traces will be accessed through the traces property which
        # updates the channellist to include only active trace numbers
        self._traces = ChannelList(self, "PNATraces", PNATrace)
        self.add_submodule("traces", self._traces)
        # Add shortcuts to first trace
        trace1 = self.traces[0]
        params = trace1.parameters
        if not isinstance(params, dict):
            raise RuntimeError(f"Expected trace.parameters to be a dict got "
                               f"{type(params)}")
        for param in params.values():
            self.parameters[param.name] = param
        # And also add a link to run sweep
        self.run_sweep = trace1.run_sweep
        # Set this trace to be the default (it's possible to end up in a
        # situation where no traces are selected, causing parameter snapshots
        # to fail)
        self.active_trace(trace1.trace_num)

        # Set auto_sweep parameter
        # If we want to return multiple traces per setpoint without sweeping
        # multiple times, we should set this to false
        self.add_parameter('auto_sweep',
                           label='Auto Sweep',
                           set_cmd=None,
                           get_cmd=None,
                           vals=Bool(),
                           initial_value=True)

        # A default output format on initialisation
        self.write('FORM REAL,32')
        self.write('FORM:BORD NORM')

        self.connect_message()

    @property
    def traces(self) -> ChannelList:
        """
        Update channel list with active traces and return the new list
        """
        # Keep track of which trace was active before. This command may fail
        # if no traces were selected.
        try:
            active_trace = self.active_trace()
        except VisaIOError as e:
            if e.error_code == errors.StatusCode.error_timeout:
                active_trace = None
            else:
                raise

        # Get a list of traces from the instrument and fill in the traces list
        parlist = self.get_trace_catalog().split(",")
        self._traces.clear()
        for trace_name in parlist[::2]:
            trace_num = self.select_trace_by_name(trace_name)
            pna_trace = PNATrace(self, f"tr{trace_num}",
                                 trace_name, trace_num)
            self._traces.append(pna_trace)

        # Restore the active trace if there was one
        if active_trace:
            self.active_trace(active_trace)

        # Return the list of traces on the instrument
        return self._traces

    def get_options(self) -> Sequence[str]:
        # Query the instrument for what options are installed
        return self.ask('*OPT?').strip('"').split(',')

    def get_trace_catalog(self) -> str:
        """
        Get the trace catalog, that is a list of trace and sweep types
        from the PNA.

        The format of the returned trace is:
            trace_name,trace_type,trace_name,trace_type...
        """
        return self.ask("CALC:PAR:CAT:EXT?").strip('"')

    def select_trace_by_name(self, trace_name: str) -> int:
        """
        Select a trace on the PNA by name.

        Returns:
            The trace number of the selected trace
        """
        self.write(f"CALC:PAR:SEL '{trace_name}'")
        return self.active_trace()

    def reset_averages(self) -> None:
        """
        Reset averaging
        """
        self.write("SENS:AVER:CLE")

    def averages_on(self) -> None:
        """
        Turn on trace averaging
        """
        self.averages_enabled(True)

    def averages_off(self) -> None:
        """
        Turn off trace averaging
        """
        self.averages_enabled(False)

    def _set_power_limits(self,
                          min_power: Union[int, float],
                          max_power: Union[int, float]) -> None:
        """
        Set port power limits
        """
        self.power.vals = Numbers(min_value=min_power,
                                  max_value=max_power)
        for port in self.ports:
            port._set_power_limits(min_power, max_power)


class PNAxBase(PNABase):
    def _enable_fom(self) -> None:
        '''
        PNA-x units with two sources have an enormous list of functions &
        configurations. In practice, most of this will be set up manually on
        the unit, with power and frequency varied in a sweep.
        '''
        self.add_parameter('aux_frequency',
                           label='Aux Frequency',
                           get_cmd='SENS:FOM:RANG4:FREQ:CW?',
                           get_parser=float,
                           set_cmd='SENS:FOM:RANG4:FREQ:CW {:.2f}',
                           unit='Hz',
                           vals=Numbers(min_value=self.min_freq,
                                        max_value=self.max_freq))



import gc

from unittest import TestCase

from qcodes.loops import Loop
from qcodes.actions import UnsafeThreadingException
from qcodes.tests.instrument_mocks import DummyInstrument


class TestUnsafeThreading(TestCase):

    def setUp(self):
        self.inst1 = DummyInstrument(name='inst1',
                                     gates=['v1', 'v2'])
        self.inst2 = DummyInstrument(name='inst2',
                                     gates=['v1', 'v2'])

    def tearDown(self):
        self.inst1.close()
        self.inst2.close()

        del self.inst1
        del self.inst2

        gc.collect()

    def test_unsafe_exception(self):
        to_meas = (self.inst1.v1, self.inst1.v2)
        loop = Loop(self.inst2.v1.sweep(0, 1, num=10)).each(*to_meas)

        with self.assertRaises(UnsafeThreadingException):
            loop.run(use_threads=True)



import warnings

import pytest
import pyvisa as visa

from qcodes.instrument.visa import VisaInstrument
from qcodes.utils.validators import Numbers


class MockVisa(VisaInstrument):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_parameter('state',
                           get_cmd='STAT?', get_parser=float,
                           set_cmd='STAT:{:.3f}',
                           vals=Numbers(-20, 20))

    def set_address(self, address):
        self.visa_handle = MockVisaHandle()
        self.visabackend = self.visalib


class MockVisaHandle(visa.resources.MessageBasedResource):
    """
    mock the API needed for a visa handle that throws lots of errors:

    - any write command sets a single "state" variable to a float
      after the last : in the command
    - a negative number results in an error raised here
    - 0 results in a return code for visa timeout
    - any ask command returns the state
    - a state > 10 throws an error
    """
    def __init__(self):
        self.state = 0
        self.closed = False

    def clear(self):
        self.state = 0

    def close(self):
        # make it an error to ask or write after close
        self.closed = True

    def write(self, cmd):
        if self.closed:
            raise RuntimeError("Trying to write to a closed instrument")
        num = float(cmd.split(':')[-1])
        self.state = num

        if num < 0:
            raise ValueError('be more positive!')

        if num == 0:
            raise visa.VisaIOError(visa.constants.VI_ERROR_TMO)

        return len(cmd)

    def ask(self, cmd):
        if self.closed:
            raise RuntimeError("Trying to ask a closed instrument")
        if self.state > 10:
            raise ValueError("I'm out of fingers")
        return self.state

    def query(self, cmd):
        if self.state > 10:
            raise ValueError("I'm out of fingers")
        return self.state

    def set_visa_attribute(
            self, name, state
    ):
        setattr(self, str(name), state)

    def __del__(self):
        pass


 # error args for set(-10)
args1 = [
    'be more positive!',
    "writing 'STAT:-10.000' to <MockVisa: Joe>",
    'setting Joe_state to -10'
]

# error args for set(0)
args2 = [
    "writing 'STAT:0.000' to <MockVisa: Joe>",
    'setting Joe_state to 0'
]

# error args for get -> 15
args3 = [
    "I'm out of fingers",
    "asking 'STAT?' to <MockVisa: Joe>",
    'getting Joe_state'
]


@pytest.fixture(name='mock_visa')
def _make_mock_visa():
    mv = MockVisa('Joe', 'none_address')
    try:
        yield mv
    finally:
        mv.close()


def test_ask_write_local(mock_visa):

    # test normal ask and write behavior
    mock_visa.state.set(2)
    assert mock_visa.state.get() == 2
    mock_visa.state.set(3.4567)
    assert mock_visa.state.get() == 3.457  # driver rounds to 3 digits

    # test ask and write errors
    with pytest.raises(ValueError) as e:
        mock_visa.state.set(-10)
    for arg in args1:
        assert arg in str(e.value)
    assert mock_visa.state.get() == -10  # set still happened

    with pytest.raises(visa.VisaIOError) as e:
        mock_visa.state.set(0)
    for arg in args2:
        assert arg in str(e.value)
    assert mock_visa.state.get() == 0

    mock_visa.state.set(15)
    with pytest.raises(ValueError) as e:
        mock_visa.state.get()
    for arg in args3:
        assert arg in str(e.value)


def test_visa_backend(mocker, request):

    rm_mock = mocker.patch('qcodes.instrument.visa.visa.ResourceManager')

    address_opened = [None]

    class MockBackendVisaInstrument(VisaInstrument):
        visa_handle = MockVisaHandle()

    class MockRM:
        def open_resource(self, address):
            address_opened[0] = address
            return MockVisaHandle()

    rm_mock.return_value = MockRM()

    inst1 = MockBackendVisaInstrument('name', address='None')
    request.addfinalizer(inst1.close)
    assert rm_mock.call_count == 1
    assert rm_mock.call_args == ((),)
    assert address_opened[0] == 'None'
    inst1.close()

    inst2 = MockBackendVisaInstrument('name2', address='ASRL2')
    request.addfinalizer(inst2.close)
    assert rm_mock.call_count == 2
    assert rm_mock.call_args == ((),)
    assert address_opened[0] == 'ASRL2'
    inst2.close()

    # this one raises a warning
    with pytest.warns(UserWarning, match="use the visalib"):
        inst3 = MockBackendVisaInstrument('name3', address='ASRL3@py')
        request.addfinalizer(inst3.close)

    assert rm_mock.call_count == 3
    assert rm_mock.call_args == (('@py',),)
    assert address_opened[0] == 'ASRL3'
    inst3.close()

    # this one doesn't
    inst4 = MockBackendVisaInstrument('name4',
                                      address='ASRL4', visalib='@py')
    request.addfinalizer(inst4.close)
    assert rm_mock.call_count == 4
    assert rm_mock.call_args == (('@py',),)
    assert address_opened[0] == 'ASRL4'
    inst4.close()


def test_visa_instr_metadata(request):
    metadatadict = {'foo': 'bar'}
    mv = MockVisa('Joe', 'none_adress', metadata=metadatadict)
    request.addfinalizer(mv.close)
    assert mv.metadata == metadatadict



import json
import re
from copy import deepcopy

import pytest

from qcodes.dataset.descriptions.param_spec import ParamSpec, ParamSpecBase
from qcodes.dataset.descriptions.dependencies import (DependencyError,
                                                      InferenceError,
                                                      InterDependencies_)
from qcodes.dataset.descriptions.versioning.v0 import InterDependencies
from qcodes.dataset.descriptions.versioning.converters import (new_to_old, old_to_new)
from qcodes.tests.common import error_caused_by



def test_wrong_input_raises():

    for pspecs in [['p1', 'p2', 'p3'],
                   [ParamSpec('p1', paramtype='numeric'), 'p2'],
                   ['p1', ParamSpec('p2', paramtype='text')]]:

        with pytest.raises(ValueError):
            InterDependencies(pspecs)


def test_init(some_paramspecbases):
    """
    Assert that the init functions correctly sets up the object.
    Assert via the public-facing methods.
    """

    (ps1, ps2, ps3, ps4) = some_paramspecbases

    idps1 = InterDependencies_(dependencies={ps1: (ps2,)})
    idps2 = InterDependencies_(dependencies={ps1: (ps2, ps2, ps2)})

    assert idps1 == idps2
    assert idps1.what_depends_on(ps2) == (ps1,)
    assert idps1.what_is_inferred_from(ps2) == ()
    assert idps1.non_dependencies == (ps1,)

    idps1 = InterDependencies_(dependencies={ps1: (ps2, ps3)})
    idps2 = InterDependencies_(dependencies={ps1: (ps3, ps2)})

    assert idps1.what_depends_on(ps2) == (ps1,)
    assert idps1.what_depends_on(ps3) == (ps1,)
    assert idps1.non_dependencies == (ps1,)
    assert idps2.non_dependencies == (ps1,)

    idps = InterDependencies_(dependencies={ps1: (ps3, ps2),
                                            ps4: (ps3,)})
    assert set(idps.what_depends_on(ps3)) == {ps1, ps4}
    assert idps.non_dependencies == (ps1, ps4)


def test_init_validation_raises(some_paramspecbases):

    (ps1, ps2, ps3, ps4) = some_paramspecbases

    # First test validation of trees invalid in their own right

    invalid_trees = ([ps1, ps2],
                     {'ps1': 'ps2'},
                     {ps1: 'ps2'},
                     {ps1: ('ps2',)},
                     {ps1: (ps2,), ps2: (ps1,)}
                     )
    causes = ("ParamSpecTree must be a dict",
              "ParamSpecTree must have ParamSpecs as keys",
              "ParamSpecTree must have tuple values",
              "ParamSpecTree can only have tuples "
              "of ParamSpecs as values",
              "ParamSpecTree can not have cycles")

    for tree, cause in zip(invalid_trees, causes):
        with pytest.raises(ValueError, match='Invalid dependencies') as ei:
            InterDependencies_(dependencies=tree, inferences={})

        assert error_caused_by(ei, cause=cause)

    for tree, cause in zip(invalid_trees, causes):
        with pytest.raises(ValueError, match='Invalid inferences') as ei:
            InterDependencies_(dependencies={}, inferences=tree)

        assert error_caused_by(ei, cause=cause)

    with pytest.raises(ValueError, match='Invalid standalones') as ei:
        InterDependencies_(standalones=('ps1', 'ps2'))

    assert error_caused_by(ei, cause='Standalones must be a sequence of '
                                     'ParamSpecs')

    # Now test trees that are invalid together

    invalid_trees = [{'deps': {ps1: (ps2, ps3)},
                      'inffs': {ps2: (ps4, ps1)}}]
    for inv in invalid_trees:
        with pytest.raises(ValueError,
                           match=re.escape("Invalid dependencies/inferences")):
            InterDependencies_(dependencies=inv['deps'],
                               inferences=inv['inffs'])

def test_to_dict(some_paramspecbases):

    def tester(idps):
        ser = idps._to_dict()
        json.dumps(ser)
        idps_deser = InterDependencies_._from_dict(ser)
        assert idps == idps_deser

    (ps1, ps2, ps3, ps4) = some_paramspecbases

    idps = InterDependencies_(standalones=(ps1, ps2),
                              dependencies={ps3: (ps4,)})
    tester(idps)

    idps = InterDependencies_(standalones=(ps1, ps2, ps3, ps4))
    tester(idps)

    idps = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                              inferences={ps2: (ps4,), ps3: (ps4,)})
    tester(idps)


def test_old_to_new_and_back(some_paramspecs):

    idps_old = InterDependencies(*some_paramspecs[1].values())
    idps_new = old_to_new(idps_old)

    assert new_to_old(idps_new) == idps_old


def test_old_to_new(some_paramspecs):

    ps1 = some_paramspecs[1]['ps1']
    ps2 = some_paramspecs[1]['ps2']
    ps3 = some_paramspecs[1]['ps3']
    ps4 = some_paramspecs[1]['ps4']
    ps5 = some_paramspecs[1]['ps5']
    ps6 = some_paramspecs[1]['ps6']

    idps_old = InterDependencies(ps1, ps2, ps3)

    idps_new = old_to_new(idps_old)

    ps1_base = ps1.base_version()
    ps2_base = ps2.base_version()
    ps3_base = ps3.base_version()
    ps4_base = ps4.base_version()
    ps5_base = ps5.base_version()
    ps6_base = ps6.base_version()


    assert idps_new.dependencies == {}
    assert idps_new.inferences == {ps3_base: (ps1_base,)}
    assert idps_new.standalones == {ps2_base}
    paramspecs = (ps1_base, ps2_base, ps3_base)
    assert idps_new._id_to_paramspec == {ps.name: ps for ps in paramspecs}

    idps_old = InterDependencies(ps2, ps4, ps1, ps2, ps3, ps5, ps6)

    idps_new = old_to_new(idps_old)

    assert idps_new.dependencies == {ps5_base: (ps3_base, ps4_base),
                                     ps6_base: (ps3_base, ps4_base)}
    assert idps_new.inferences == {ps3_base: (ps1_base,),
                                   ps4_base: (ps2_base,)}
    assert idps_new.standalones == set()
    paramspecs = (ps1_base, ps2_base, ps3_base, ps4_base, ps5_base, ps6_base)
    assert idps_new._id_to_paramspec == {ps.name: ps for ps in paramspecs}

    idps_old = InterDependencies(ps1, ps2)

    idps_new = old_to_new(idps_old)

    assert idps_new.dependencies == {}
    assert idps_new.inferences == {}
    assert idps_new.standalones == {ps1_base, ps2_base}
    paramspecs = (ps1_base, ps2_base)
    assert idps_new._id_to_paramspec == {ps.name: ps for ps in paramspecs}


def test_new_to_old(some_paramspecbases):

    (ps1, ps2, ps3, ps4) = some_paramspecbases

    idps_new = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                                  standalones=(ps4,))

    paramspec1 = ParamSpec(name=ps1.name, paramtype=ps1.type,
                           label=ps1.label, unit=ps1.unit,
                           depends_on=[ps2.name, ps3.name])
    paramspec2 = ParamSpec(name=ps2.name, paramtype=ps2.type,
                           label=ps2.label, unit=ps2.unit)
    paramspec3 = ParamSpec(name=ps3.name, paramtype=ps3.type,
                           label=ps3.label, unit=ps3.unit)
    paramspec4 = ParamSpec(name=ps4.name, paramtype=ps4.type,
                           label=ps4.label, unit=ps4.unit)
    idps_old_expected = InterDependencies(paramspec2, paramspec3,
                                          paramspec1, paramspec4)

    assert new_to_old(idps_new) == idps_old_expected

    #

    idps_new = InterDependencies_(inferences={ps1: (ps2, ps3)},
                                  standalones=(ps4,))

    paramspec1 = ParamSpec(name=ps1.name, paramtype=ps1.type,
                           label=ps1.label, unit=ps1.unit,
                           inferred_from=[ps2.name, ps3.name])
    paramspec2 = ParamSpec(name=ps2.name, paramtype=ps2.type,
                           label=ps2.label, unit=ps2.unit)
    paramspec3 = ParamSpec(name=ps3.name, paramtype=ps3.type,
                           label=ps3.label, unit=ps3.unit)
    paramspec4 = ParamSpec(name=ps4.name, paramtype=ps4.type,
                           label=ps4.label, unit=ps4.unit)
    idps_old_expected = InterDependencies(paramspec2, paramspec3,
                                          paramspec1, paramspec4)

    assert new_to_old(idps_new) == idps_old_expected



def test_extend_with_paramspec(some_paramspecs):
    ps1 = some_paramspecs[1]['ps1']
    ps2 = some_paramspecs[1]['ps2']
    ps3 = some_paramspecs[1]['ps3']
    ps4 = some_paramspecs[1]['ps4']
    ps5 = some_paramspecs[1]['ps5']
    ps6 = some_paramspecs[1]['ps6']


    ps1_base = ps1.base_version()
    ps2_base = ps2.base_version()
    ps3_base = ps3.base_version()
    ps4_base = ps4.base_version()
    ps5_base = ps5.base_version()
    ps6_base = ps6.base_version()

    idps_bare = InterDependencies_(standalones=(ps1_base,))
    idps_extended = InterDependencies_(inferences={ps3_base: (ps1_base,)})

    assert idps_bare._extend_with_paramspec(ps3) == idps_extended

    idps_bare = InterDependencies_(standalones=(ps2_base,),
                                   inferences={ps3_base: (ps1_base,)})
    idps_extended = InterDependencies_(inferences={ps3_base: (ps1_base,),
                                                   ps4_base: (ps2_base,)})

    assert idps_bare._extend_with_paramspec(ps4) == idps_extended

    idps_bare = InterDependencies_(standalones=(ps1_base, ps2_base))
    idps_extended = InterDependencies_(
                        inferences={ps3_base: (ps1_base,),
                                    ps4_base: (ps2_base,)},
                        dependencies={ps5_base: (ps3_base, ps4_base),
                                      ps6_base: (ps3_base, ps4_base)})
    assert (idps_bare.
            _extend_with_paramspec(ps3).
            _extend_with_paramspec(ps4).
            _extend_with_paramspec(ps5).
            _extend_with_paramspec(ps6)) == idps_extended


def test_validate_subset(some_paramspecbases):

    ps1, ps2, ps3, ps4 = some_paramspecbases

    idps = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                              inferences={ps2: (ps4,), ps3: (ps4,)})

    idps.validate_subset((ps4,))
    idps.validate_subset((ps2, ps4))
    idps.validate_subset((ps2, ps3, ps4))
    idps.validate_subset(())
    idps.validate_subset([])

    with pytest.raises(DependencyError) as exc_info:
        idps.validate_subset((ps1,))
    assert exc_info.value._param_name == 'psb1'
    assert exc_info.value._missing_params == {'psb2', 'psb3'}

    with pytest.raises(DependencyError) as exc_info:
        idps.validate_subset((ps1, ps2, ps4))
    assert exc_info.value._param_name == 'psb1'
    assert exc_info.value._missing_params == {'psb3'}

    with pytest.raises(InferenceError) as exc_info:
        idps.validate_subset((ps3,))
    assert exc_info.value._param_name == 'psb3'
    assert exc_info.value._missing_params == {'psb4'}

    with pytest.raises(InferenceError) as exc_info:
        idps2 = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                                    inferences={ps3: (ps4,)})
        idps2.validate_subset((ps1, ps2, ps3))
    assert exc_info.value._param_name == 'psb3'
    assert exc_info.value._missing_params == {'psb4'}

    with pytest.raises(ValueError, match='ps42'):
        ps42 = ParamSpecBase('ps42', paramtype='text', label='', unit='it')
        idps.validate_subset((ps2, ps42, ps4))


def test_extend(some_paramspecbases):

    ps1, ps2, ps3, _ = some_paramspecbases

    idps = InterDependencies_(standalones=(ps1, ps2))

    idps_ext = idps.extend(dependencies={ps1: (ps3,)})
    idps_expected = InterDependencies_(standalones=(ps2,),
                                       dependencies={ps1: (ps3,)})
    assert idps_ext == idps_expected

    # lazily check that we get brand new objects
    idps._id_to_paramspec[ps1.name].label = "Something new and awful"
    idps._id_to_paramspec[ps2.name].unit = "Ghastly unit"
    assert idps_ext._id_to_paramspec[ps1.name].label == 'blah'
    assert idps_ext._id_to_paramspec[ps2.name].unit == 'V'
    # reset the objects that are never supposed to be mutated
    idps._id_to_paramspec[ps1.name].label = "blah"
    idps._id_to_paramspec[ps2.name].unit = "V"

    idps = InterDependencies_(standalones=(ps2,))
    idps_ext = idps.extend(dependencies={ps1: (ps2,)})
    idps_expected = InterDependencies_(dependencies={ps1: (ps2,)})
    assert idps_ext == idps_expected

    idps = InterDependencies_(dependencies={ps1: (ps2,)})
    idps_ext = idps.extend(dependencies={ps1: (ps2, ps3)})
    idps_expected = InterDependencies_(dependencies={ps1: (ps2, ps3)})
    assert idps_ext == idps_expected

    idps = InterDependencies_()
    idps_ext = idps.extend(standalones=(ps1, ps2))
    idps_expected = InterDependencies_(standalones=(ps2, ps1))
    assert idps_ext == idps_expected

    ps_nu = deepcopy(ps1)
    ps_nu.unit += '/s'
    idps = InterDependencies_(standalones=(ps1,))
    idps_ext = idps.extend(standalones=(ps_nu,))
    idps_expected = InterDependencies_(standalones=(ps_nu, ps1))
    assert idps_ext == idps_expected

    idps = InterDependencies_(dependencies={ps1: (ps2,)})
    match = re.escape("Invalid dependencies/inferences")
    with pytest.raises(ValueError, match=match):
        idps_ext = idps.extend(inferences={ps2: (ps1,)})


def test_remove(some_paramspecbases):
    ps1, ps2, ps3, ps4 = some_paramspecbases

    idps = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                              inferences={ps2: (ps4, )})
    idps_rem = idps.remove(ps1)
    idps_expected = InterDependencies_(inferences={ps2: (ps4,)},
                                       standalones=(ps3,))
    assert idps_rem == idps_expected

    for p in [ps4, ps2, ps3]:
        match = re.escape(f'Cannot remove {p.name}, other parameters')
        with pytest.raises(ValueError, match=match):
            idps_rem = idps.remove(p)

    idps = InterDependencies_(dependencies={ps1: (ps3,)},
                              inferences={ps2: (ps4,)})
    idps_rem = idps.remove(ps2)
    idps_expected = InterDependencies_(dependencies={ps1: (ps3,)},
                                       standalones=(ps4,))

    assert idps_rem == idps_expected

    idps = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                              standalones=(ps4, ))
    idps_rem = idps.remove(ps4)
    idps_expected = InterDependencies_(dependencies={ps1: (ps2, ps3)})
    assert idps_rem == idps_expected

    idps = InterDependencies_(dependencies={ps1: (ps2, ps3)},
                              standalones=(ps4, ))
    idps_rem = idps.remove(ps1)
    idps_expected = InterDependencies_(standalones=(ps2, ps3, ps4))
    assert idps_rem == idps_expected

def test_equality_old(some_paramspecs):

    # TODO: make this more fancy with itertools

    ps1 = some_paramspecs[1]['ps1']
    ps2 = some_paramspecs[1]['ps2']
    ps3 = some_paramspecs[1]['ps3']
    ps4 = some_paramspecs[1]['ps4']
    ps5 = some_paramspecs[1]['ps5']
    ps6 = some_paramspecs[1]['ps6']

    assert InterDependencies(ps1, ps2, ps3) == InterDependencies(ps3, ps2, ps1)
    assert InterDependencies(ps1, ps6, ps3) == InterDependencies(ps3, ps6, ps1)
    assert InterDependencies(ps4, ps5, ps3) == InterDependencies(ps3, ps4, ps5)


def test_non_dependents():
    ps1 = ParamSpecBase('ps1', paramtype='numeric', label='Raw Data 1',
                        unit='V')
    ps2 = ParamSpecBase('ps2', paramtype='array', label='Raw Data 2',
                        unit='V')
    ps3 = ParamSpecBase('ps3', paramtype='text', label='Axis 1',
                        unit='')
    ps4 = ParamSpecBase('ps4', paramtype='numeric', label='Axis 2',
                        unit='V')
    ps5 = ParamSpecBase('ps5', paramtype='numeric', label='Signal',
                        unit='Conductance')
    ps6 = ParamSpecBase('ps6', paramtype='text', label='Goodness',
                        unit='')

    idps1 = InterDependencies_(dependencies={ps5: (ps3, ps4), ps6: (ps3, ps4)},
                               inferences={ps4: (ps2,), ps3: (ps1,)})

    assert idps1.non_dependencies == (ps5, ps6)

    idps2 = InterDependencies_(dependencies={ps2: (ps1,)})

    assert idps2.non_dependencies == (ps2,)

    idps3 = InterDependencies_(dependencies={ps6: (ps1,)},
                               standalones=(ps2,))

    assert idps3.non_dependencies == (ps2, ps6)
